import unittest
import logging
from ..feeps_tools.mms_read_feeps_sector_masks_csv import mms_read_feeps_sector_masks_csv
from pyspedas import mms_load_feeps, mms_feeps_pad
from pyspedas.projects.mms.feeps_tools.mms_feeps_gpd import mms_feeps_gpd
from pyspedas import del_data, tplot, data_exists, get

global_display=False

class FEEPSTestCases(unittest.TestCase):
    def test_time_clip_regression(self):
        # regression test for time clipping bug with spin-averaged spectra
        mms_load_feeps(trange=['2015-12-15/10:00', '2015-12-15/12:00'], time_clip=True)
        data = get('mms1_epd_feeps_srvy_l2_electron_intensity_omni_spin')
        self.assertTrue(data.y[-1, :].sum() != 0.0)

    def test_log_filtering(self):
        # Ensure that log message filtering can be reliably enabled and disabled
        # Enable filtering
        saw_filtered_entry = False
        with self.assertLogs(level='WARNING') as cm:
            # Guarantee at least one warning is logged
            logging.warning("Dummy log entry")
            mms_load_feeps(trange=['2015-12-15/10:00', '2015-12-15/12:00'], filter_recvary_warnings=True)
            for log_output in cm.output:
                if  'record-varying' in log_output:
                    saw_filtered_entry = True
        self.assertFalse(saw_filtered_entry)
        # Disable filtering, ensure that filter from previous call was successfully removed
        saw_filtered_entry = False
        with self.assertLogs(level='WARNING') as cm:
            # Guarantee at least one warning is logged
            logging.warning("Dummy log entry")
            mms_load_feeps(trange=['2015-12-15/10:00', '2015-12-15/12:00'], filter_recvary_warnings=False)
            for log_output in cm.output:
                if 'record-varying' in log_output:
                    saw_filtered_entry = True
        self.assertTrue(saw_filtered_entry)

    def test_feeps_sitl(self):
        mms_load_feeps(datatype='electron', trange=['2016-11-23', '2016-11-24'], data_rate='srvy', probe=4,
                       level='sitl')
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_sitl_electron_intensity_omni'))

    def test_feeps_pad_regression(self):
        """
        This is a regression test for a bug caused by the v7 of the FEEPS CDF files
        v7 CDFs have an extra dimension on the variables containing the pitch angle data
        """
        mms_load_feeps(datatype='electron', trange=['2016-11-23', '2016-11-24'], data_rate='srvy', probe=4)
        mms_feeps_pad(probe=4)
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad'))
        tplot('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad', display=global_display)
        del_data('*')

    def test_gyrophase_angles(self):
        mms_load_feeps(data_rate='brst', trange=['2017-07-11/22:34', '2017-07-11/22:34:25'], probe=3)
        mms_feeps_gpd(probe='3', energy=[61, 77], data_rate='brst')
        self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_61-77keV_gpd'))
        mms_feeps_gpd(probe='3', data_rate='brst')
        self.assertTrue(data_exists('mms3_epd_feeps_brst_l2_electron_intensity_50-500keV_gpd'))
        tplot(['mms3_epd_feeps_brst_l2_electron_intensity_61-77keV_gpd',
               'mms3_epd_feeps_brst_l2_electron_intensity_50-500keV_gpd'], display=global_display)

    def test_pad_ions_brst(self):
        mms_load_feeps(probe=4, data_rate='brst', datatype='ion', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        mms_feeps_pad(probe=4, data_rate='brst', datatype='ion', angles_from_bfield=True)
        self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad'))
        self.assertTrue(data_exists('mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad_spin'))
        tplot(['mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad',
               'mms4_epd_feeps_brst_l2_ion_intensity_70-600keV_pad_spin'], display=global_display)

    def test_pad_ions_srvy(self):
        mms_load_feeps(probe=4, datatype='ion', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        mms_feeps_pad(probe=4, datatype='ion')
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad'))
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad_spin'))
        tplot(['mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad',
               'mms4_epd_feeps_srvy_l2_ion_intensity_70-600keV_pad_spin'], display=global_display)

    def test_pad_electrons_srvy(self):
        mms_load_feeps()
        mms_feeps_pad()
        self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'))
        self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad'))
        tplot(['mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad',
               'mms1_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'], display=global_display)

    def test_pad_electrons_srvy_probe(self):
        mms_load_feeps(probe=4)
        mms_feeps_pad(probe=4)
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'))
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad'))
        tplot(['mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad',
               'mms4_epd_feeps_srvy_l2_electron_intensity_70-600keV_pad_spin'], display=global_display)

    def test_electron_srvy_after_aug17(self):
        # there's a different set of active eyes after 16 August 2017
        # this test executes that code
        mms_load_feeps(probe=4, trange=['2017-12-01', '2017-12-02'])
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_omni'))
        self.assertTrue(data_exists('mms4_epd_feeps_srvy_l2_electron_intensity_omni_spin'))
        tplot(['mms4_epd_feeps_srvy_l2_electron_intensity_omni',
               'mms4_epd_feeps_srvy_l2_electron_intensity_omni_spin'], display=global_display)

    def test_feeps_pct_err(self):
        mms_load_feeps(trange = ['2020-08-03/01:05:00', '2020-08-03/01:08:00'],
                        probe = '2',
                        level = 'l1b',
                        data_rate = 'brst',
                        datatype = 'electron',
                        data_units = 'counts')

        feeps_vars = mms_load_feeps(trange = ['2020-08-03/01:05:00','2020-08-03/01:08:00'],
                   probe = '2',
                   level = 'l2',
                   data_rate = 'brst',
                   datatype = 'electron',
                   data_units = 'intensity',
                   get_err = True)
        self.assertTrue(data_exists('mms2_epd_feeps_brst_l2_electron_intensity_omni_percent_uncertainty'))
        self.assertTrue('mms2_epd_feeps_brst_l2_electron_intensity_omni_percent_uncertainty' in feeps_vars)
        tplot(['mms2_epd_feeps_brst_l2_electron_intensity_omni',
            'mms2_epd_feeps_brst_l2_electron_intensity_omni_percent_uncertainty'],
            display=global_display,
            save_png='feeps_percent_uncertainty.png')

    def test_sector_masks(self):
        d = mms_read_feeps_sector_masks_csv(['2015-08-01', '2015-08-02'])
        self.assertTrue(d['mms4imaskt2'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms3imaskb3'] == [18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48])
        self.assertTrue(d['mms4imaskt8'] == [0, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 57, 58, 59, 61])
        self.assertTrue(d['mms4imaskb4'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms4imaskt1'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms1imaskt2'] == [11, 12])
        self.assertTrue(d['mms3imaskb1'] == [20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43])
        self.assertTrue(d['mms3imaskb5'] == [36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61])
        self.assertTrue(d['mms1imaskt3'] == [22, 23, 24, 25, 26, 27, 28, 31, 32])
        d = mms_read_feeps_sector_masks_csv(['2016-08-01', '2016-08-02'])
        self.assertTrue(d['mms1imaskb4'] == [20, 21, 22, 23, 24, 25, 26, 27, 28, 37, 38, 47])
        self.assertTrue(d['mms4imaskb9'] == [31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 53, 54, 55, 56])
        self.assertTrue(d['mms3imaskb9'] == [53, 54, 55, 56, 57])
        self.assertTrue(d['mms4imaskb7'] == [30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43])
        self.assertTrue(d['mms3imaskb1'] == [20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43])
        self.assertTrue(d['mms4imaskt9'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 25, 26, 27, 28, 29, 30, 31, 32, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63])
        self.assertTrue(d['mms3imaskb8'] == [31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45])
        d = mms_read_feeps_sector_masks_csv(['2017-08-01', '2017-08-02'])
        self.assertTrue(d['mms3imaskb9'] == [53, 54, 55, 56, 57, 58])
        self.assertTrue(d['mms2imaskt9'] == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 53, 54, 60, 61, 62, 63])
        self.assertTrue(d['mms1imaskb6'] == [40, 41, 42, 49, 50, 51, 52, 53, 54, 57, 58])

    def test_all_sector_masks(self):
        # Ensure that all contamination sector CSV files are readable, with no trailing commas
        # or other formatting issues.  Update the file_times array in this test whenever new
        # CSVs are added.

        from pyspedas.tplot_tools import time_string

        file_times=[1447200000.0000000, # 11/11/2015
             1468022400.0000000, # 7/9/2016
             1477612800.0000000, # 10/28/2016
             1496188800.0000000, # 5/31/2017
             1506988800.0000000, # 10/3/2017
             1538697600.0000000, # 10/5/2018
             1642032000.0000000, # 1/13/2022
             1651795200.0000000, # 5/6/2022
             1660521600.0000000, # 8/15/2022
             1706832000.0000000, # 02/02/2024
             1721779200.0000000, # 07/24/2024
             1739664000.0000000, # 02/16/2025
             1755129600.0000000] # 08/14/2025

        for ft in file_times:
            trange=[ft,ft+86400.0]
            logging.info("Reading files for: "+time_string(trange[0]))
            d = mms_read_feeps_sector_masks_csv(trange)
            self.assertTrue(d is not None)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pyspedas import clip, options, store_data, ylim, zlim, get_data

from ..load import load
from ..get_gatt_ror import get_gatt_ror

from typing import List, Optional

def xep(
    trange: List[str] = ['2017-06-01', '2017-06-02'],
    datatype: str = 'omniflux',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the XEP-e experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-06-01', '2017-06-02']

        datatype: str
            Data type; Valid options: 'omniflux', '2dflux'
            Default: 'omniflux'

        level: str
            Data level; Valid options: 'l2'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.  Default: None

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded.  Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        ror: bool
            If set, print PI info and rules of the road
            Default: True

        uname: str
            User name. Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> xep_vars = pyspedas.projects.erg.xep(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_xep_l2_FEDO_SSD')

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    if (datatype == 'omniflux') or (datatype == '2dflux'):
        # to avoid failure of creation Tplot variables (at store_data.py) of xep
        notplot = True
    file_res = 3600. * 24
    prefix = 'erg_xep_'+level+'_'
    pathformat = 'satellite/erg/xep/'+level+'/'+datatype + \
        '/%Y/%m/erg_xep_'+level+'_'+datatype+'_%Y%m%d_v??_??.cdf'
    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)
            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG XEP')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ",gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print('RoR of XEP: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Xep')
            print('')
            print('Contact: erg_xep_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if isinstance(loaded_data, dict):

        if datatype == 'omniflux':
            tplot_variables = []

            if prefix + 'FEDO_SSD' + suffix in loaded_data:
                v_vars_min = loaded_data[prefix + 'FEDO_SSD' + suffix]['v'][0]
                v_vars_max = loaded_data[prefix + 'FEDO_SSD' + suffix]['v'][1]
                v_vars = np.sqrt(v_vars_min * v_vars_max)  # Geometric mean

                store_data(prefix + 'FEDO_SSD' + suffix, data={'x': loaded_data[prefix + 'FEDO_SSD' + suffix]['x'],
                                                               'y': loaded_data[prefix + 'FEDO_SSD' + suffix]['y'],
                                                               'v': v_vars},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDO_SSD' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDO_SSD' + suffix)

                if prefix + 'FEDO_SSD' + suffix in tplot_variables:
                    # remove minus valuse of y array
                    clip(prefix + 'FEDO_SSD' + suffix, 0., 5000.)
                # set spectrogram plot option
                options(prefix + 'FEDO_SSD' + suffix, 'Spec', 1)
                # set y axis to logscale
                options(prefix + 'FEDO_SSD' + suffix, 'ylog', 1)
                # set yrange
                options(prefix + 'FEDO_SSD' + suffix,
                        'yrange', [4.0e+02, 4.5e+03])
                # set z axis to logscale
                options(prefix + 'FEDO_SSD' + suffix, 'zlog', 1)
                # set zrange
                options(prefix + 'FEDO_SSD' + suffix,
                        'zrange', [1.0e-01, 1.0e+3])
                # change colormap option
                options(prefix + 'FEDO_SSD' + suffix, 'Colormap', 'jet')

                # set ztitle
                options(prefix + 'FEDO_SSD' + suffix,
                        'ztitle', '[/cm^{2}-str-s-keV]')
                # set ytitle
                options(prefix + 'FEDO_SSD' + suffix,
                        'ytitle', 'XEP\nomniflux\nLv2\nEnergy')
                # set ysubtitle
                options(prefix + 'FEDO_SSD' + suffix, 'ysubtitle', '[keV]')

                ylim(prefix + 'FEDO_SSD' + suffix, 4.0e+02, 4.5e+03)
                zlim(prefix + 'FEDO_SSD' + suffix, 1.0e-01, 1.0e+3)

            return tplot_variables

        if datatype == '2dflux':
            tplot_variables = []

            if prefix + 'FEDU_SSD' + suffix in loaded_data:
                v_keyname = 'v'
                if v_keyname not in loaded_data[prefix + 'FEDU_SSD' + suffix]:
                    v_keyname = 'v1'
                store_data(prefix + 'FEDU_SSD' + suffix,
                           data={'x': loaded_data[prefix + 'FEDU_SSD' + suffix]['x'],
                                 'y': loaded_data[prefix + 'FEDU_SSD' + suffix]['y'],
                                 'v1': np.sqrt(loaded_data[prefix + 'FEDU_SSD' + suffix][v_keyname][:9, 0]
                                               * loaded_data[prefix + 'FEDU_SSD' + suffix][v_keyname][:9, 1]),  # Geometric mean of 'v'
                                 'v2': [i for i in range(16)]},  # [0, 1, 2, .., 15]
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_SSD' + suffix]['CDF']})

                tplot_variables.append(prefix + 'FEDU_SSD' + suffix)

                if prefix + 'FEDU_SSD' + suffix in tplot_variables:
                    clip(prefix + 'FEDU_SSD' + suffix, -1.0e+10, 1.0e+10)

                return tplot_variables

    return loaded_data



import numpy as np
import unittest
from pyspedas import xdegap


class TestXdegap(unittest.TestCase):
    def test_no_nan_values(self):
        """Test array with no NaN values"""
        input_array = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        result = xdegap(input_array)
        np.testing.assert_array_equal(result, input_array)

    def test_single_nan_middle(self):
        """Test array with single NaN value in the middle"""
        input_array = np.array([1.0, 2.0, np.nan, 4.0, 5.0])
        expected = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        result = xdegap(input_array)
        np.testing.assert_array_almost_equal(result, expected)

    def test_multiple_consecutive_nans(self):
        """Test array with multiple consecutive NaN values"""
        input_array = np.array([1.0, np.nan, np.nan, 4.0, 5.0])
        # Naive expectation: multiple nans will be linearly interpolated to
        # the array indices:
        # expected = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        # Actual behavior: Pairwise averaging,left to right, of surrounding non-nan values
        # 2.5 is halfway between 1.0 and 4.0
        # 3.25 is halfway between 2.5 and 4.0
        expected = np.array([1.0, 2.5, 3.25, 4.0, 5.0])
        result = xdegap(input_array)
        np.testing.assert_array_almost_equal(result, expected)

    def test_nan_at_start(self):
        """Test array with NaN at the start"""
        input_array = np.array([np.nan, 2.0, 3.0, 4.0, 5.0])
        expected = np.array([2.0, 2.0, 3.0, 4.0, 5.0])
        result = xdegap(input_array)
        np.testing.assert_array_almost_equal(result, expected)

    def test_nan_at_end(self):
        """Test array with NaN at the end"""
        input_array = np.array([1.0, 2.0, 3.0, 4.0, np.nan])
        expected = np.array([1.0, 2.0, 3.0, 4.0, 4.0])
        result = xdegap(input_array)
        np.testing.assert_array_almost_equal(result, expected)

    def test_all_nan(self):
        """Test array with all NaN values"""
        input_array = np.array([np.nan, np.nan, np.nan])
        result = xdegap(input_array)
        self.assertTrue(np.all(np.isnan(result)))


if __name__ == "__main__":
    unittest.main()



import unittest
from pyspedas.tplot_tools import data_exists, del_data
import pyspedas


class EquatorS_Tests(unittest.TestCase):
    def test_load_notplot(self):
        mam_vars = pyspedas.projects.equator_s.mam(notplot=True)
        self.assertTrue("B_xyz_gse%eq_pp_mam" in mam_vars)

    def test_load_mam_data(self):
        mam_vars = pyspedas.projects.equator_s.mam(time_clip=True)
        self.assertTrue(data_exists("B_xyz_gse%eq_pp_mam"))

    def test_load_aux_data(self):
        aux_vars = pyspedas.projects.equator_s.load(instrument="aux", prefix="test_")
        self.assertTrue(data_exists("test_sc_r_xyz_gse%eq_pp_aux"))

    def test_load_edi_data(self):
        edi_vars = pyspedas.projects.equator_s.edi(downloadonly=True)
        edi_vars = pyspedas.projects.equator_s.edi(no_update=True)
        self.assertTrue(data_exists("V_ed_xyz_gse%eq_pp_edi"))
        self.assertTrue(data_exists("E_xyz_gse%eq_pp_edi"))

    def test_load_edi_sp(self):
        edi_vars = pyspedas.projects.equator_s.edi(datatype="sp")
        self.assertTrue(data_exists("V_ed_xyz_gse%eq_sp_edi"))

    def test_load_epi_data(self):
        epi_vars = pyspedas.projects.equator_s.epi()
        self.assertTrue(data_exists("J_e_1%eq_pp_epi"))
        self.assertTrue(data_exists("J_e_2%eq_pp_epi"))

    def test_load_ici_data(self):
        ici_vars = pyspedas.projects.equator_s.ici()
        self.assertTrue(data_exists("N_p%eq_pp_ici"))

    def test_load_pcd_data(self):
        pcd_vars = pyspedas.projects.equator_s.pcd()
        self.assertTrue(data_exists("I_ion%eq_pp_pcd"))

    def test_load_sfd_data(self):
        sfd_vars = pyspedas.projects.equator_s.sfd(trange=["1998-01-26", "1998-01-27"])
        self.assertTrue(data_exists("F_e>0.26%eq_sp_sfd"))

    def test_load_all_data(self):
        all_vars = pyspedas.projects.equator_s.load(instrument="all")
        self.assertTrue(data_exists("J_e_1%eq_pp_epi"))
        self.assertTrue(data_exists("J_e_2%eq_pp_epi"))

    def test_load_invalid_trange(self):
        del_data()
        t_all = pyspedas.projects.equator_s.load(trange="bad input")
        self.assertTrue(len(t_all) == 0)
        del_data()
        t_all = pyspedas.projects.equator_s.load(trange=["2018-12-15"])
        self.assertTrue(len(t_all) == 0)
        del_data()
        t_all = pyspedas.projects.equator_s.load(trange=["2018-12-15", "2018-12-14"])
        self.assertTrue(len(t_all) == 0)

    def test_load_invalid_instrument(self):
        del_data()
        t_all = pyspedas.projects.equator_s.load(instrument="bad input")
        self.assertTrue(len(t_all) == 0)


if __name__ == "__main__":
    unittest.main()



from .epd import epd_load as epd
from .state import state_load as state
from .mrmi import mrmi_load as mrmi
from .mrma import mrma_load as mrma
from .fgm import fgm_load as fgm
from .eng import eng_load as eng



import numpy as np
import logging
from pyspedas.utilities.interpol import interpol


def interp_gap(x, y):
    """
    Replace NaNs with interpolated values for the y array.
    If there are any NaNs in the x array, no interpolation is performed.

    This function replaces NaN values in y with interpolated values based on x.
    For multidimensional y arrays, interpolation is performed on each column.

    Parameters
    ----------
    x : array_like
        Independent variable array (e.g., time). Must be 1-dimensional
        and finite.
    y : array_like
        Dependent variable array to interpolate. Can be 1D or 2D.

    Returns
    -------
    dict
        Dictionary containing:
            'y': Modified y array with NaNs replaced by interpolated values
            'index': Indices where interpolation was performed
            'count': Number of points that were interpolated

    Notes
    -----
    This function mirrors the behavior of the IDL procedure interp_gap.
    It uses the interpol function for the actual interpolation.

    Examples
    --------
    >>> import numpy as np
    >>> from pyspedas.utilities.interp_gap import interp_gap
    >>> x = np.array([1, 2, 3, 4, 5])
    >>> y = np.array([10, np.nan, 30, np.nan, 50])
    >>> result = interp_gap(x, y)
    >>> print(result['y'])  # [10, 20, 30, 40, 50]
    """

    # Convert inputs to numpy arrays
    x = np.asarray(x)
    y = np.asarray(y)

    # Check that x values are finite
    invalid_x = ~np.isfinite(x)
    if np.any(invalid_x):
        logging.error("Invalid x points")
        return {"y": y, "index": [-1], "count": 0}

    # Handle multidimensional case (operate per column and union indices)
    if y.ndim > 1:
        index = np.zeros(len(x), dtype=bool)
        y_modified = y.copy()

        for i in range(y.shape[1]):
            column = y[:, i].copy()
            result = interp_gap(x, column)
            y_modified[:, i] = result["y"]
            if result["count"] > 0:
                index[result["index"]] = True

        wb = np.where(index)[0]
        return {"y": y_modified, "index": wb, "count": len(wb)}

    # Handle 1D case
    ny = len(y)

    # Find NaN/infinite values in y
    wb = np.where(~np.isfinite(y))[0]
    count = len(wb)

    if count == 0:
        # No bad points
        return {"y": y, "index": np.array([]), "count": 0}

    logging.info("Found %d out of %d bad points", count, ny)

    # Find neighboring good points for interpolation
    wbp_candidates = np.concatenate([wb - 1, wb + 1])
    wbp_candidates = np.clip(wbp_candidates, 0, ny - 1)
    wbp = np.unique(wbp_candidates)

    # Keep only finite neighboring points in y
    finite_mask = np.isfinite(y[wbp])
    if not np.any(finite_mask):
        logging.info("All points bad, no interpolation done")
        return {"y": y, "index": wb, "count": count}

    wbp = wbp[finite_mask]
    xwbp = x[wbp]
    ywbp = y[wbp]

    # Extend with boundary points (flat extrapolation at edges)
    x_extended = np.concatenate(([x[0]], xwbp, [x[ny - 1]]))
    y_extended = np.concatenate(([ywbp[0]], ywbp, [ywbp[-1]]))

    # Interpolate at bad point locations
    try:
        yp = interpol(y_extended, x_extended, x[wb])
        y_result = y.copy()
        y_result[wb] = yp
        return {"y": y_result, "index": wb, "count": count}
    except (ValueError, IndexError) as e:
        logging.info("Interpolation failed: %s", str(e))
        return {"y": y, "index": wb, "count": count}



from .load import load
from pyspedas import options
from pyspedas.utilities.datasets import find_datasets

# This routine was originally in ace/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.

def sis(trange=['2018-11-5', '2018-11-6'],
        datatype='k0',
        prefix='',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        force_download=False):
    """
    Load data from the ACE Solar Isotope Spectrometer (SIS)
    
    Parameters
    ----------

        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-11-5', '2018-11-6']

        datatype: str
            Data type; Valid options::

                h1: 256-sec Level 2 Data
                h2: 1-Hour Level 2 Data
                k0: 1-Hour Key Parameters

            Default: 'k0'

        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
            Default: None

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)
            Default: []

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    --------
        list of str
            A list of the tplot variables created.

    Examples
    ----------

        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> sis_vars = pyspedas.projects.ace.sis(trange=['2018-11-5', '2018-11-6'])
        >>> tplot(['H_lo', 'H_hi', 'CNO_lo', 'CNO_hi', 'Z_ge_10'])

    """
    return load(trange=trange, instrument='sis', datatype=datatype, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                varformat=varformat, downloadonly=downloadonly, notplot=notplot, no_update=no_update, varnames=varnames,
                time_clip=time_clip, force_download=force_download)





from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(trange=None,
         probe='1A',
         datatype='sspc', 
         level='l2',
         version='v10',
         get_support_data=False,
         files='',
         notplot=False,
         prefix='',
         suffix='',
         downloadonly=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the BARREL mission; this function is not meant 
    to be called directly; instead, see the wrappers::

        pyspedas.projects.barrel.sspc
        pyspedas.projects.barrel.mspc
        pyspedas.projects.barrel.fspc
        pyspedas.projects.barrel.magn
        pyspedas.projects.barrel.ephm
        pyspedas.projects.barrel.rcnt
        pyspedas.projects.barrel.hkpg

    """

    if not isinstance(probe, list):
        probe = [probe]
    
    if trange is None:
        trange = CONFIG['defaults'][probe[0]]

    if prefix is None:
        prefix = ''
    if suffix is None:
        suffix = ''

    out_files = []
    for prb in probe:
        if str(datatype) == 'ephm':
            # SPDF has 'ephem' instead of 'ephm' for this data type
            remote_path = (str(level) + '/' + str(prb) + '/' + 'ephem' +
                           '/bar_' + str(prb) + '_' + str(level) + '_' +
                           str(datatype) + '_%Y%m%d_' + str(version) + '.cdf')
        else:
            remote_path = (str(level) + '/' + str(prb) + '/' + str(datatype) +
                           '/bar_' + str(prb) + '_' + str(level) + '_' + str(datatype) +
                           '_%Y%m%d_' + str(version) + '.cdf')

        remote_names = [name.lower() for name in dailynames(file_format=remote_path, trange=trange)]

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'],
                         local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)
    
    if downloadonly:
        return out_files

    # Convert the cdf files to tvars.
    # Make sure each of the variables is prefixed with the flight ID
    tvars = []
    for file in out_files:
        p_start = file.find("bar_")
        p_end = file.find("_", p_start + len("bar_"))
        flight = str.upper(file[p_start+len("bar_"):p_end + 1])
        internal_prefix = "brl"+flight
        prefix = prefix+internal_prefix
        tvars = tvars + cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data, notplot=notplot)

    if len(tvars) == 0:
        return

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars




import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data, tplot_names, get_data
from pyspedas.projects.erg.satellite.erg.particle.erg_lepi_get_dist import erg_lepi_get_dist
from pyspedas.projects.erg.satellite.erg.particle.erg_mepi_get_dist import erg_mepi_get_dist
from pyspedas.projects.erg.satellite.erg.particle.erg_lepe_get_dist import erg_lepe_get_dist
from pyspedas.projects.erg.satellite.erg.particle.erg_mepe_get_dist import erg_mepe_get_dist
from pyspedas.projects.erg.satellite.erg.particle.erg_hep_get_dist import erg_hep_get_dist
from pyspedas.projects.erg.satellite.erg.particle.erg_xep_get_dist import erg_xep_get_dist
import numpy as np
import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_load_att_data(self):
        del_data()
        att_vars = pyspedas.projects.erg.att()
        tplot_names()
        self.assertTrue(data_exists('erg_att_sprate'))
        self.assertTrue(data_exists('erg_att_spphase'))
        self.assertTrue('erg_att_sprate' in att_vars)
        self.assertTrue('erg_att_spphase' in att_vars)

    def test_load_att_data_notplot(self):
        del_data()
        att_dict = pyspedas.projects.erg.att(notplot=True)
        self.assertTrue('erg_att_sprate' in att_dict)
        self.assertTrue('erg_att_spphase' in att_dict)


    def test_load_hep_data(self):
        del_data()
        hep_vars = pyspedas.projects.erg.hep()
        self.assertTrue(data_exists('erg_hep_l2_FEDO_L'))
        self.assertTrue(data_exists('erg_hep_l2_FEDO_H'))
        self.assertTrue('erg_hep_l2_FEDO_L' in hep_vars)
        self.assertTrue('erg_hep_l2_FEDO_H' in hep_vars)

    def test_load_hep_3dflux(self):
        del_data()
        hep_vars = pyspedas.projects.erg.hep(datatype='3dflux')
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L'))
        self.assertTrue(data_exists('erg_hep_l2_FEDU_H'))
        self.assertTrue('erg_hep_l2_FEDU_L' in hep_vars)
        self.assertTrue('erg_hep_l2_FEDU_H' in hep_vars)

    def test_load_hep_l3_data(self):
        del_data()
        hep_vars = pyspedas.projects.erg.hep(level='l3')
        self.assertTrue(data_exists('erg_hep_l3_FEDU_L_paspec_ene00'))
        self.assertTrue(data_exists('erg_hep_l3_FEDU_H'))
        self.assertTrue('erg_hep_l3_FEDU_L_paspec_ene00' in hep_vars)
        self.assertTrue('erg_hep_l3_FEDU_H' in hep_vars)


    def test_load_xep_l3_data(self):
        del_data()
        # There are no XEP L3 products yet, but we can at least get some code coverage
        xep_vars = pyspedas.projects.erg.xep(level='l3')

    def test_load_xep_l2_omniflux_data(self):
        del_data()
        xep_vars = pyspedas.projects.erg.xep(level='l2')
        self.assertTrue('erg_xep_l2_FEDO_SSD' in xep_vars)
        self.assertTrue(data_exists('erg_xep_l2_FEDO_SSD'))


    def test_load_xep_l2_2dflux_data(self):
        del_data()
        xep_vars = pyspedas.projects.erg.xep(level='l2', datatype='2dflux')
        self.assertTrue('erg_xep_l2_FEDU_SSD' in xep_vars)
        self.assertTrue(data_exists('erg_xep_l2_FEDU_SSD'))

    def test_load_orb_data(self):
        del_data()
        orb_vars = pyspedas.projects.erg.orb()
        self.assertTrue(data_exists('erg_orb_l2_pos_gse'))
        self.assertTrue(data_exists('erg_orb_l2_pos_gsm'))
        self.assertTrue(data_exists('erg_orb_l2_pos_sm'))
        self.assertTrue(data_exists('erg_orb_l2_vel_gse'))
        self.assertTrue(data_exists('erg_orb_l2_vel_gsm'))
        self.assertTrue(data_exists('erg_orb_l2_vel_sm'))
        self.assertTrue('erg_orb_l2_pos_gse' in orb_vars)
        self.assertTrue('erg_orb_l2_pos_gsm' in orb_vars)
        self.assertTrue('erg_orb_l2_pos_sm' in orb_vars)
        self.assertTrue('erg_orb_l2_vel_gse' in orb_vars)
        self.assertTrue('erg_orb_l2_vel_gsm' in orb_vars)
        self.assertTrue('erg_orb_l2_vel_sm' in orb_vars)

    def test_load_orb_pre_data(self):
        del_data()
        orb_vars = pyspedas.projects.erg.orb(datatype='pre')
        self.assertTrue(data_exists('erg_orb_pre_l2_pos_gse'))
        self.assertTrue(data_exists('erg_orb_pre_l2_pos_gsm'))
        self.assertTrue(data_exists('erg_orb_pre_l2_pos_sm'))
        self.assertTrue(data_exists('erg_orb_pre_l2_vel_gse'))
        self.assertTrue(data_exists('erg_orb_pre_l2_vel_gsm'))
        self.assertTrue(data_exists('erg_orb_pre_l2_vel_sm'))
        self.assertTrue('erg_orb_pre_l2_pos_gse' in orb_vars)
        self.assertTrue('erg_orb_pre_l2_pos_gsm' in orb_vars)
        self.assertTrue('erg_orb_pre_l2_pos_sm' in orb_vars)
        self.assertTrue('erg_orb_pre_l2_vel_gse' in orb_vars)
        self.assertTrue('erg_orb_pre_l2_vel_gsm' in orb_vars)
        self.assertTrue('erg_orb_pre_l2_vel_sm' in orb_vars)

    def test_load_l3_orb_data(self):
        del_data()
        orb_vars = pyspedas.projects.erg.orb(level='l3')
        self.assertTrue(data_exists('erg_orb_l3_pos_eq_op'))
        self.assertTrue(data_exists('erg_orb_l3_pos_iono_north_op'))
        self.assertTrue(data_exists('erg_orb_l3_pos_iono_south_op'))
        self.assertTrue('erg_orb_l3_pos_eq_op' in orb_vars)
        self.assertTrue('erg_orb_l3_pos_iono_north_op' in orb_vars)
        self.assertTrue('erg_orb_l3_pos_iono_south_op' in orb_vars)

    def test_load_mgf_data(self):
        del_data()
        mgf_vars = pyspedas.projects.erg.mgf(time_clip=True)
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_sm'))
        self.assertTrue('erg_mgf_l2_mag_8sec_sm' in mgf_vars)

    def test_load_lepe_data_v4(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe(version='v04_01')
        self.assertTrue(data_exists('erg_lepe_l2_omniflux_FEDO'))
        self.assertTrue('erg_lepe_l2_omniflux_FEDO' in lepe_vars)

    def test_load_lepe_data_v5(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe()
        self.assertTrue(data_exists('erg_lepe_l2_omniflux_FEDO'))
        self.assertTrue('erg_lepe_l2_omniflux_FEDO' in lepe_vars)

    def test_load_lepe_3dflux_v4(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe(datatype='3dflux', version='v04_01')
        self.assertTrue('erg_lepe_l2_3dflux_FEDU' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_FEDU'))
        self.assertTrue('erg_lepe_l2_3dflux_Count_Rate' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_Count_Rate'))
        self.assertTrue('erg_lepe_l2_3dflux_Count_Rate_BG' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_Count_Rate_BG'))
        fedu_dat = get_data('erg_lepe_l2_3dflux_FEDU')
        cr_dat = get_data('erg_lepe_l2_3dflux_Count_Rate')
        bg_dat = get_data('erg_lepe_l2_3dflux_Count_Rate_BG')
        fedu_dims = fedu_dat[1].shape
        cr_dims = cr_dat[1].shape
        bg_dims = bg_dat[1].shape
        self.assertEqual(fedu_dims, (9962,32,12,16))
        self.assertEqual(cr_dims, (9962,32,12,16))
        self.assertEqual(bg_dims, (9962,32,16))

    def test_load_lepe_3dflux_v5(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe(datatype='3dflux')
        self.assertTrue('erg_lepe_l2_3dflux_FEDU' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_FEDU'))
        self.assertTrue('erg_lepe_l2_3dflux_Count_Rate' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_Count_Rate'))
        self.assertTrue('erg_lepe_l2_3dflux_Count_Rate_BG' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_Count_Rate_BG'))
        fedu_dat = get_data('erg_lepe_l2_3dflux_FEDU')
        cr_dat = get_data('erg_lepe_l2_3dflux_Count_Rate')
        bg_dat = get_data('erg_lepe_l2_3dflux_Count_Rate_BG')
        fedu_dims = fedu_dat[1].shape
        cr_dims = cr_dat[1].shape
        bg_dims = bg_dat[1].shape
        # The v5 CDFs contain fewer data points than the v4 CDFs (quality filtering?)
        self.assertEqual(fedu_dims, (9552,32,12,16))
        self.assertEqual(cr_dims, (9552,32,12,16))
        self.assertEqual(bg_dims, (9552,32,16))


    def test_load_lepe_l3_data_v5(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe(level='l3', et_diagram=True)
        self.assertTrue(data_exists('erg_lepe_l3_pa_FEDU'))
        self.assertTrue('erg_lepe_l3_pa_FEDU' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l3_pa_enech_01_FEDU'))
        self.assertTrue('erg_lepe_l3_pa_enech_01_FEDU' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l3_pa_pabin_01_FEDU'))
        self.assertTrue('erg_lepe_l3_pa_pabin_01_FEDU' in lepe_vars)

    def test_load_lepe_l3_data_v4(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe(level='l3', et_diagram=True, version='v04_01')
        self.assertTrue(data_exists('erg_lepe_l3_pa_FEDU'))
        self.assertTrue('erg_lepe_l3_pa_FEDU' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l3_pa_enech_01_FEDU'))
        self.assertTrue('erg_lepe_l3_pa_enech_01_FEDU' in lepe_vars)
        self.assertTrue(data_exists('erg_lepe_l3_pa_pabin_01_FEDU'))
        self.assertTrue('erg_lepe_l3_pa_pabin_01_FEDU' in lepe_vars)

    def test_load_lepi_data(self):
        del_data()
        lepi_vars = pyspedas.projects.erg.lepi()
        self.assertTrue(data_exists('erg_lepi_l2_omniflux_FPDO'))
        self.assertTrue(data_exists('erg_lepi_l2_omniflux_FHEDO'))
        self.assertTrue(data_exists('erg_lepi_l2_omniflux_FODO'))
        self.assertTrue('erg_lepi_l2_omniflux_FPDO' in lepi_vars)
        self.assertTrue('erg_lepi_l2_omniflux_FHEDO' in lepi_vars)
        self.assertTrue('erg_lepi_l2_omniflux_FODO' in lepi_vars)

    def test_load_lepi_3dflux_data(self):
        del_data()
        lepi_vars = pyspedas.projects.erg.lepi(datatype='3dflux')
        self.assertTrue(data_exists('erg_lepi_l2_3dflux_FPDU'))
        self.assertTrue(data_exists('erg_lepi_l2_3dflux_FHEDU'))
        self.assertTrue('erg_lepi_l2_3dflux_FPDU' in lepi_vars)
        self.assertTrue('erg_lepi_l2_3dflux_FHEDU' in lepi_vars)


    def test_load_mepi_get_dist(self):
        del_data()
        mepi_vars = pyspedas.projects.erg.mepi_nml(datatype='3dflux')
        base = "erg_mepi_l2_3dflux_"
        for species in ['FPDU', 'FHEDU', 'FODU']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_mepi_get_dist(var, single_time="2017-07-01 00:00:00", species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_mepi_get_dist(var, index=[0], species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_mepi_get_dist(var, index=0, species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_mepi_get_dist(var, index=[0,1,2], species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_mepi_get_dist(var, index=np.array([0,1,2]), species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_mepi_get_dist(var, trange=[t[0,], t[-1]], species=None)
            self.assertTrue(len(dist['time']) == len(t))

        self.assertTrue(data_exists('erg_mepi_l2_3dflux_FPDU'))
        self.assertTrue(data_exists('erg_mepi_l2_3dflux_FHEDU'))
        self.assertTrue('erg_mepi_l2_3dflux_FPDU' in mepi_vars)
        self.assertTrue('erg_mepi_l2_3dflux_FHEDU' in mepi_vars)

    def test_load_lepi_get_dist(self):
        del_data()
        lepi_vars = pyspedas.projects.erg.lepi(datatype='3dflux')
        base = "erg_lepi_l2_3dflux_"
        for species in ['FPDU', 'FHEDU', 'FODU']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_lepi_get_dist(var, single_time="2017-07-01 00:00:00", species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepi_get_dist(var, index=[0], species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepi_get_dist(var, index=0, species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepi_get_dist(var, index=[0,1,2], species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_lepi_get_dist(var, index=np.array([0,1,2]), species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_lepi_get_dist(var, trange=[t[0,], t[-1]], species=None)
            self.assertTrue(len(dist['time']) == len(t))

        self.assertTrue(data_exists('erg_lepi_l2_3dflux_FPDU'))
        self.assertTrue(data_exists('erg_lepi_l2_3dflux_FHEDU'))
        self.assertTrue('erg_lepi_l2_3dflux_FPDU' in lepi_vars)
        self.assertTrue('erg_lepi_l2_3dflux_FHEDU' in lepi_vars)

    def test_load_lepe_get_dist_v4(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe(datatype='3dflux',version='v04_01')
        base = "erg_lepe_l2_3dflux_"
        for species in ['FEDU']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_lepe_get_dist(var, single_time="2017-04-04 00:00:00", species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepe_get_dist(var, index=[0], species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepe_get_dist(var, index=0, species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepe_get_dist(var, index=[0,1,2], species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_lepe_get_dist(var, index=np.array([0,1,2]), species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_lepe_get_dist(var, trange=[t[0,], t[-1]], species=None)
            self.assertTrue(len(dist['time']) == len(t))
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_FEDU'))
        self.assertTrue('erg_lepe_l2_3dflux_FEDU' in lepe_vars)

    def test_load_lepe_get_dist_v5(self):
        del_data()
        lepe_vars = pyspedas.projects.erg.lepe(datatype='3dflux')
        base = "erg_lepe_l2_3dflux_"
        for species in ['FEDU']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_lepe_get_dist(var, single_time="2017-04-04 00:00:00", species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepe_get_dist(var, index=[0], species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepe_get_dist(var, index=0, species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_lepe_get_dist(var, index=[0,1,2], species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_lepe_get_dist(var, index=np.array([0,1,2]), species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_lepe_get_dist(var, trange=[t[0,], t[-1]], species=None)
            self.assertTrue(len(dist['time']) == len(t))
        self.assertTrue(data_exists('erg_lepe_l2_3dflux_FEDU'))
        self.assertTrue('erg_lepe_l2_3dflux_FEDU' in lepe_vars)

    def test_load_mepe_get_dist(self):
        del_data()
        mepe_vars = pyspedas.projects.erg.mepe(datatype='3dflux')
        base = "erg_mepe_l2_3dflux_"
        for species in ['FEDU']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_mepe_get_dist(var, single_time="2017-03-27 00:00:00", species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_mepe_get_dist(var, index=[0], species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_mepe_get_dist(var, index=0, species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_mepe_get_dist(var, index=[0,1,2], species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_mepe_get_dist(var, index=np.array([0,1,2]), species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_mepe_get_dist(var, trange=[t[0,], t[-1]], species=None)
            self.assertTrue(len(dist['time']) == len(t))
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU'))
        self.assertTrue('erg_mepe_l2_3dflux_FEDU' in mepe_vars)

    def test_load_hep_get_dist(self):
        del_data()
        hep_vars = pyspedas.projects.erg.hep(datatype='3dflux')
        base = "erg_hep_l2_"
        for species in ['FEDU_L', 'FEDU_H']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_hep_get_dist(var, single_time="2017-03-27 00:00:00", species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_hep_get_dist(var, index=[0], species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_hep_get_dist(var, index=0, species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_hep_get_dist(var, index=[0,1,2], species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_hep_get_dist(var, index=np.array([0,1,2]), species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_hep_get_dist(var, trange=[t[0,], t[-1]], species=None)
            # This assertion doesn't necessarily hold for HEP.  Something to do with
            # availability of support data?
            # self.assertTrue(len(dist['time']) == len(t))
        self.assertTrue(data_exists('erg_hep_l2_FEDU_L'))
        self.assertTrue('erg_hep_l2_FEDU_L' in hep_vars)

    def test_load_xep_get_dist(self):
        del_data()
        xep_vars = pyspedas.projects.erg.xep(datatype='2dflux')
        base = "erg_xep_l2_"
        for species in ['FEDU_SSD']:
            var = base + species
            vardat = get_data(var)
            varmeta = get_data(var, metadata=True)
            t = vardat[0]
            dist = erg_xep_get_dist(var, single_time="2017-06-01 00:00:00", species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_xep_get_dist(var, index=[0], species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_xep_get_dist(var, index=0, species=None)
            self.assertTrue(len(dist['time']) == 1)
            dist = erg_xep_get_dist(var, index=[0,1,2], species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_xep_get_dist(var, index=np.array([0,1,2]), species=None)
            self.assertTrue(len(dist['time']) == 3)
            dist = erg_xep_get_dist(var, trange=[t[0,], t[-1]], species=None)
            # This assertion doesn't necessarily hold for HEP.  Something to do with
            # availability of support data?
            # self.assertTrue(len(dist['time']) == len(t))
        self.assertTrue(data_exists('erg_xep_l2_FEDU_SSD'))
        self.assertTrue('erg_xep_l2_FEDU_SSD' in xep_vars)


    def test_load_lepi_l3_data(self):
        del_data()
        lepi_vars = pyspedas.projects.erg.lepi(level='l3')
        self.assertTrue(data_exists('erg_lepi_l3_pa_FPDU'))
        self.assertTrue('erg_lepi_l3_pa_FPDU' in lepi_vars)
        self.assertTrue(data_exists('erg_lepi_l3_pa_pabin_01_FPDU'))
        self.assertTrue('erg_lepi_l3_pa_pabin_01_FPDU' in lepi_vars)

    def test_load_mepe_data(self):
        del_data()
        mepe_vars = pyspedas.projects.erg.mepe()
        self.assertTrue(data_exists('erg_mepe_l2_omniflux_FEDO'))
        self.assertTrue('erg_mepe_l2_omniflux_FEDO' in mepe_vars)

    def test_load_mepe_l3_data(self):
        del_data()
        mepe_vars = pyspedas.projects.erg.mepe(level='l3')
        self.assertTrue(data_exists('erg_mepe_l3_3dflux_FEDU'))
        self.assertTrue('erg_mepe_l3_3dflux_FEDU' in mepe_vars)

    def test_load_mepe_l3_pa_data(self):
        del_data()
        mepe_vars = pyspedas.projects.erg.mepe(level='l3', datatype='pa')
        self.assertTrue(data_exists('erg_mepe_l3_pa_FEDU'))
        self.assertTrue('erg_mepe_l3_pa_FEDU' in mepe_vars)

    def test_load_mepe_3dflux(self):
        del_data()
        mepe_vars = pyspedas.projects.erg.mepe(datatype='3dflux')
        self.assertTrue(data_exists('erg_mepe_l2_3dflux_FEDU'))
        self.assertTrue('erg_mepe_l2_3dflux_FEDU' in mepe_vars)

    def test_load_mepi_nml_data(self):
        del_data()
        mepi_vars = pyspedas.projects.erg.mepi_nml()
        self.assertTrue(data_exists('erg_mepi_l2_omniflux_epoch_tof'))
        self.assertTrue('erg_mepi_l2_omniflux_epoch_tof' in mepi_vars)

    def test_load_mepi_nml_3dflux_data(self):
        del_data()
        mepi_vars = pyspedas.projects.erg.mepi_nml(datatype='3dflux')
        self.assertTrue(data_exists('erg_mepi_l2_3dflux_FPDU'))
        self.assertTrue('erg_mepi_l2_3dflux_FPDU' in mepi_vars)

    def test_load_mepi_tof_data(self):
        del_data()
        mepi_vars = pyspedas.projects.erg.mepi_tof()
        self.assertTrue(data_exists('erg_mepi_l2_tofflux_FPDU'))
        self.assertTrue(data_exists('erg_mepi_l2_tofflux_FODU'))
        self.assertTrue('erg_mepi_l2_tofflux_FPDU' in mepi_vars)
        self.assertTrue('erg_mepi_l2_tofflux_FODU' in mepi_vars)

    def test_load_mepi_nml_l3_3dflux_data(self):
        del_data()
        mepi_vars = pyspedas.projects.erg.mepi_nml(level='l3', datatype='3dflux')
        self.assertTrue(data_exists('erg_mepi_l3_3dflux_FPDU'))
        self.assertTrue('erg_mepi_l3_3dflux_FPDU' in mepi_vars)

    def test_load_mepi_nml_l3_pa_data(self):
        del_data()
        mepi_vars = pyspedas.projects.erg.mepi_nml(level='l3', datatype='pa')
        self.assertTrue(data_exists('erg_mepi_l3_pa_FPDU'))
        self.assertTrue('erg_mepi_l3_pa_FPDU' in mepi_vars)


    def test_load_pwe_ofa_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_ofa()
        self.assertTrue(data_exists('erg_pwe_ofa_l2_spec_E_spectra_132'))
        self.assertTrue(data_exists('erg_pwe_ofa_l2_spec_B_spectra_132'))
        self.assertTrue('erg_pwe_ofa_l2_spec_E_spectra_132' in pwe_vars)
        self.assertTrue('erg_pwe_ofa_l2_spec_B_spectra_132' in pwe_vars)

    def test_load_pwe_efd_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_efd()
        self.assertTrue(data_exists('erg_pwe_efd_l2_E_spin_Eu_dsi'))
        self.assertTrue('erg_pwe_efd_l2_E_spin_Eu_dsi' in pwe_vars)

    def test_load_pwe_efd_e64hz_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_efd(datatype='E64Hz')
        self.assertTrue(data_exists('erg_pwe_efd_l2_E64Hz_dsi_Ex_waveform'))
        self.assertTrue('erg_pwe_efd_l2_E64Hz_dsi_Ex_waveform' in pwe_vars)

    def test_load_pwe_efd_pot_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_efd(datatype='pot')
        self.assertTrue(data_exists('erg_pwe_efd_l2_pot_Vu1'))
        self.assertTrue('erg_pwe_efd_l2_pot_Vu1' in pwe_vars)

    def test_load_pwe_efd_spec_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_efd(datatype='spec')
        self.assertTrue(data_exists('erg_pwe_efd_l2_spec_spectra'))
        self.assertTrue('erg_pwe_efd_l2_spec_spectra' in pwe_vars)

    def test_load_pwe_hfa_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_hfa()
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_eu'))
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_ev'))
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_esum'))
        self.assertTrue(data_exists('erg_pwe_hfa_l2_low_spectra_er'))
        self.assertTrue('erg_pwe_hfa_l2_low_spectra_eu' in pwe_vars)
        self.assertTrue('erg_pwe_hfa_l2_low_spectra_ev' in pwe_vars)
        self.assertTrue('erg_pwe_hfa_l2_low_spectra_esum' in pwe_vars)
        self.assertTrue('erg_pwe_hfa_l2_low_spectra_er' in pwe_vars)

    def test_load_pwe_wfc_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_wfc(trange=['2017-04-01/12:00:00', '2017-04-01/13:00:00'])
        self.assertTrue(data_exists('erg_pwe_wfc_l2_e_65khz_Ex_waveform'))
        self.assertTrue('erg_pwe_wfc_l2_e_65khz_Ex_waveform' in pwe_vars)

    def test_load_pwe_wfc_spec_data(self):
        del_data()
        pwe_vars = pyspedas.projects.erg.pwe_wfc(trange=['2017-04-01/12:00:00', '2017-04-01/13:00:00'], datatype='spec')
        self.assertTrue(data_exists('erg_pwe_wfc_l2_e_65khz_E_spectra'))
        self.assertTrue('erg_pwe_wfc_l2_e_65khz_E_spectra' in pwe_vars)

    def test_downloadonly(self):
        del_data()
        files = pyspedas.projects.erg.mgf(downloadonly=True, trange=['2017-03-27', '2017-03-28'])
        self.assertTrue(os.path.exists(files[0]))

if __name__ == '__main__':
    unittest.main()


import os

CONFIG = {
    "local_data_dir": "mica_data/",
    "remote_data_dir": "http://mirl.unh.edu/ULF/cdf/",
}

# override local data directory with environment variables
if os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join([os.environ["SPEDAS_DATA_DIR"], "mica"])

if os.environ.get("MICA_DATA_DIR"):
    CONFIG["local_data_dir"] = os.environ["MICA_DATA_DIR"]



from pyspedas import tinterpol
from pyspedas.tplot_tools import get_data


def mms_pgs_clean_support(times, mag_name=None, vel_name=None, sc_pot_name=None):
    """
    Transform and/or interpolate support data to match the particle data

    Parameters
    ----------
        mag_name: str
            Tplot variable containing magnetic field data

        vel_name: str
            Tplot variable containing bulk velocity data

        sc_pot_name: str
            Tplot variable containing spacecraft potential data

    Returns
    ----------
        Tuple containing interpolated (magnetic field, velocity, spacecraft potential)
    """
    out_mag = None
    out_vel = None
    out_scpot = None

    if mag_name is not None:
        mag_temp = mag_name + '_pgs_temp'
        tinterpol(mag_name, times, newname=mag_temp)
        interpolated_bfield = get_data(mag_temp)
        if interpolated_bfield is not None:
            out_mag = interpolated_bfield.y

    if vel_name is not None:
        vel_temp = vel_name + '_pgs_temp'
        tinterpol(vel_name, times, newname=vel_temp)
        interpolated_vel = get_data(vel_temp)
        if interpolated_vel is not None:
            out_vel = interpolated_vel.y

    if sc_pot_name is not None:
        scpot_temp = sc_pot_name + '_pgs_temp'
        tinterpol(sc_pot_name, times, newname=scpot_temp)
        interpolated_scpot = get_data(scpot_temp)
        if interpolated_scpot is not None:
            out_scpot = interpolated_scpot.y

    return out_mag, out_vel, out_scpot



import logging
import numpy as np
from pyspedas.tplot_tools import store_data, options

from pyspedas.particles.spd_units_string import spd_units_string


def spd_pgs_make_tplot(name, x=None, y=None, z=None, units='', ylog=False, zlog=True, colorbar='spedas', ytitle=None, ysubtitle=''):
    """
    Create tplot variable with standard spectrogram settings

    Parameters
    ----------
        name: str
            Name of the new tplot variable to create

        x: numpy.ndarray
            X-axis values (time)

        y: numpy.ndarray
            Y-axis values

        z: numpy.ndarray
            Z-axis values (data)

        units: str
            Units string to store in the metadata

        ylog: bool
            Set the y-axis to log scale (default: False)

        zlog: bool
            Set the z-axis to log scale (default: True)

        colorbar: str
            PyTplot 'Colormap' option (default: 'spedas')

    Returns
    -------
    str
        String containing new variable name

    """

    if not isinstance(x, np.ndarray) or not isinstance(y, np.ndarray) or not isinstance(z, np.ndarray) :
        logging.error('Error, must specify x, y and z parameters')
        return

    if ytitle is None:
        ytitle = name

    store_data(name, data={'x': x, 'y': z, 'v': y})
    options(name, 'ylog', ylog)
    options(name, 'zlog', zlog)
    options(name, 'Spec', True)
    options(name, 'ytitle', ytitle)
    options(name, 'ysubtitle', ysubtitle)
    options(name, 'ztitle', spd_units_string(units, units_only=True))
    options(name, 'Colormap', colorbar)
    return name



import os

CONFIG = {
    "local_data_dir": "cdaweb/",
    "cdas_endpoint": "https://cdaweb.gsfc.nasa.gov/WS/cdasr/1/dataviews/sp_phys/",
    "remote_data_dir": "https://cdaweb.gsfc.nasa.gov/sp_phys/data",
}

# override local data directory with environment variables
if os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join([os.environ["SPEDAS_DATA_DIR"], "cdaweb"])



import os
import datetime
import logging
import numpy as np
import xarray as xr
import pyspedas
from pyspedas.tplot_tools import store_data
import copy
import time


def read_column_names(sts_file):
    """
    Opens an STS file and reads the column names and vector names.

    Parameters
    ----------
    sts_file : str
        The full path to the STS file.

    Returns
    -------
    tuple of list of str
        A tuple containing two lists. The first list contains the column names, and the second list contains the vector names.

    Notes
    -----
        The .sts file format is a text file format used by the MAVEN mission.
        It is a simple format that contains a header section with metadata and a data section with the actual data.
        This function reads the header section to extract the column names and vector names of the data section.

    """
    column_names = []
    vec_names = []
    record_object = False
    in_scalar_and_vector_object = False
    in_vector_object = False
    in_scalar_object = False

    with open(sts_file, "r") as f:
        for line in f:
            line = line.split()
            if line == ["OBJECT", "=", "RECORD"]:
                record_object = True
            if record_object:
                if line == ["OBJECT", "=", "VECTOR"]:
                    in_vector_object = True
                    vec_column_names = []
                if line == ["OBJECT", "=", "SCALAR"]:
                    if in_vector_object:
                        in_scalar_and_vector_object = True
                    else:
                        in_scalar_object = True
                if line == ["END_OBJECT"]:
                    if in_scalar_and_vector_object:
                        in_scalar_and_vector_object = False
                    elif in_vector_object:
                        in_vector_object = False
                        column_names.append(vec_column_names)
                    elif in_scalar_object:
                        in_scalar_object = False
                    else:
                        # This is where we exit
                        return column_names, vec_names
                if line[0] == "NAME":
                    if in_vector_object and not in_scalar_and_vector_object:
                        vec_name = line[-1]
                        vec_names.append(vec_name)
                    if in_scalar_and_vector_object:
                        column_name = vec_name + "_" + line[-1]
                        vec_column_names.append(column_name)
                    if in_scalar_object:
                        column_name = line[-1]
                        column_names.append([column_name])
                        vec_names.append(None)


def sts_to_tplot(
    filenames=None,
    prefix="",
    suffix="",
    merge=False,
    notplot=False,
):
    """
    Convert STS files to tplot variables.

    Parameters
    ----------
    filenames : str or list of str, optional
        The full path(s) to the STS file(s). If a single string is provided, it will be treated as a single file path.
        If a list of strings is provided, each string will be treated as a separate file path.
        Default is None.
    prefix : str, optional
        A prefix to be added to the variable names when creating tplot variables.
        Default is an empty string.
    suffix : str, optional
        A suffix to be added to the variable names when creating tplot variables.
        Default is an empty string.
    merge : bool, optional
        If True, then data from 'filenames' will be merged into existing tplot variables.
        If False (default), then data from 'filenames' will overwrite existing tplot variables.
        Data in 'filenames' will always be merged/combined by themselves.
    notplot : bool, optional
        If True, the STS data will be returned as a dictionary without creating tplot variables.
        If False, tplot variables will be created.
        Default is False.

    Returns
    -------
    list of str or dict
        Returns a list of the tplot variables created.
        If notplot is True, a dictionary containing the STS data is returned.

    Notes
    -----
        The .sts file format is a text file format used by the MAVEN mission.
        It is a simple format that contains a header section with metadata and a data section with the actual data.
        This function reads the data section of the STS file and converts it to tplot variables.

    Examples
    --------
        >>> f = ["/maven/data/sci/mag/l2/2015/01/mvn_mag_l2_2015002ss1s_20150102_v01_r01.sts",
                 "/maven/data/sci/mag/l2/2015/01/mvn_mag_l2_2015001ss1s_20150101_v01_r01.sts"]
        >>> vars = sts_to_tplot(filenames=f)
        >>> print(vars)
        ['DDAY', 'OB_N', 'POSN', 'OB_BDPL']
    """

    # Create a dictionary and list in which we'll store STS variable data and variable names, respectively
    start_t = time.time()
    sts_dict = {}  # this dictionary will store the STS headers and data
    stored_variables = []

    if prefix is None:
        prefix = ""
    if suffix is None:
        suffix = ""

    # Code assumes a list of STS files
    if isinstance(filenames, str):
        filenames = [filenames]
    elif isinstance(filenames, list):
        filenames = filenames
    else:
        logging.error("Invalid filenames input.")
        return stored_variables
    filenames.sort()

    for s_file in filenames:
        column_names, vec_names = read_column_names(s_file)
        headers = [
            item for sublist in column_names for item in sublist
        ]  # a list of headers for the data
        with open(s_file, "r") as f:
            lines = f.readlines()
        # In STS files, the beginning of the data starts after the last time 'END_OBJECT' is found
        end_objects = [l for l, line in enumerate(lines) if "END_OBJECT" in line]
        end_headers = end_objects[-1]
        data = lines[end_headers + 1 :]
        data = [
            d.strip().split() for d in data
        ]  # Remove extra spaces, then split on whitespaces
        # Create the STS dictionary
        for h, head in enumerate(headers):
            data_column = [d[h] for d in data]
            if head not in sts_dict:
                # Store the data column in the dictionary
                sts_dict[head] = data_column
            else:
                # If the header already exists, append the data column to the existing data
                sts_dict[head].extend(data_column)

    # At this point we read all the files and have the data in the sts_dict dictionary

    # We need to create datetime objects from the sts_dict's year, doy, hour, min, sec, and msec data
    year = sts_dict["TIME_YEAR"]
    doy = sts_dict["TIME_DOY"]
    hour = sts_dict["TIME_HOUR"]
    min = sts_dict["TIME_MIN"]
    sec = sts_dict["TIME_SEC"]
    msec = sts_dict["TIME_MSEC"]

    dtimes = [
        datetime.datetime(
            int(yr),
            1,
            1,
            int(hr),
            int(mn),
            int(s),
            int(ms) * 1000,
            tzinfo=datetime.timezone.utc,
        )
        + datetime.timedelta(int(dy) - 1)
        for yr, dy, hr, mn, s, ms in zip(year, doy, hour, min, sec, msec)
    ]

    sts_dict["time_unix"] = [dt.timestamp() for dt in dtimes]
    # These keys are no longer necessary, remove them
    remove_time_keys = [
        "TIME_YEAR",
        "TIME_DOY",
        "TIME_HOUR",
        "TIME_MIN",
        "TIME_SEC",
        "TIME_MSEC",
    ]
    for key in remove_time_keys:
        try:
            sts_dict.pop(key)
        except KeyError:
            logging.info("Key {} was not found".format(key))

    # Don't create tplot vars if that's not what's desired
    if notplot:
        # Return a dictionary of the STS data
        return sts_dict

    # Create tplot variables
    for cn, vn in zip(column_names, vec_names):
        to_merge = False
        var_name = ""
        if vn == "TIME" or cn[0] == "DDAY":
            # We already have a time_unix key in the dictionary
            continue
        if vn is None:
            # Scalar variables
            var_name = prefix + cn[0] + suffix
            var_data = [
                float(c) for c in sts_dict[cn[0]]
            ]  # dict contains strings, convert to floats
        else:
            # Vector variables
            var_name = prefix + vn + suffix
            var_cn = [
                c for c in cn if "RANGE" not in c and "range" not in c
            ]  # do not include RANGE columns
            var_data = [
                [float(sts_dict[c][i]) for c in var_cn]
                for i in range(len(sts_dict[cn[0]]))
            ]  # data has as many dimensions as len(var_cn)

        # Check if we need to merge with existing tplot variables
        if var_name in pyspedas.tplot_tools.data_quants.keys() and merge:
            prev_data_quant = pyspedas.tplot_tools.data_quants[var_name]
            to_merge = True

        # Store the new data
        store_data(
            var_name,
            data={"x": sts_dict["time_unix"], "y": var_data},
        )
        stored_variables.append(var_name)

        # If merging is needed, merge the new data with the existing data
        if to_merge is True:
            cur_data_quant = pyspedas.tplot_tools.data_quants[var_name]
            plot_options = copy.deepcopy(pyspedas.tplot_tools.data_quants[var_name].attrs)
            pyspedas.tplot_tools.data_quants[var_name] = xr.concat(
                [prev_data_quant, cur_data_quant], dim="time"
            ).sortby("time")
            pyspedas.tplot_tools.data_quants[var_name].attrs = plot_options

    return stored_variables



"""Test plotting functions (mostly for pseudovariables)"""

import os
import unittest
import numpy as np
from pyspedas.projects import themis, elfin, fast, erg, maven, mms, psp
from pyspedas import (
    store_data,
    options,
    timespan,
    tplot,
    tplot_options,
    degap,
    del_data,
    databar,
    ylim,
    xlim,
    zlim,
    time_double,
    timebar,
    highlight,
    get_data,
    tplot_copy,
    split_vec,
    count_traces,
    annotate,
    is_pseudovariable,
)
from pyspedas.utilities.config_testing import TESTING_CONFIG

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
save_dir = os.path.join(output_dir, "utilities")
if not os.path.exists(save_dir):
    os.makedirs(save_dir)

default_trange = ["2007-03-23", "2007-03-24"]


class PlotTestCases(unittest.TestCase):
    """Test plot functions."""

    def test_markers_and_symbols(self):
        # Regression test for lineplot crash when marker sizes are set
        # Taken from pytplot markers and symbols notebook in pyspedas_examples
        del_data("*")
        store_data("data", data={"x": [1, 2, 3, 4, 5, 6], "y": [1, 1, 1, 1, 1, 1]})
        timespan("1970-01-01", 10, "seconds")
        tplot_options("title", "Simple line plot")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "simple_lineplot.png"))
        options("data", "marker", "X")
        tplot_options("title", "Simple line plot with X markers")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "markers_lineplot.png"))
        options("data", "linestyle", "None")
        tplot_options("title", "Symbol-only plot with X markers")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "noline_lineplot.png"))
        options("data", "marker_size", 200)
        tplot_options("title", "Symbol-only plot with size 200 markers")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "markersize_lineplot.png"))
        options("data", "line_style_name", "solid")
        tplot_options("title", "Line+symbol plot with size 200 markers")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "markersize_nosymbols_lineplot.png"))
        options("data", "marker_size", 20)
        tplot_options("title", "Line+symbol plot with error bars and size 20 markers")
        tplot(
            "data",
            display=global_display,
            save_png=os.path.join(save_dir, "markersize20_nosymbols_lineplot.png"),
        )
        options("data", "markevery", 2)
        tplot_options("title", "Line+symbol plot with error bars and markers every 2 data points")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "markevery_lineplot.png"))
        options("data", "marker", "H")
        tplot_options("title", "Line plot with error bars and hexagon markers")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "hexagons_lineplot.png"))
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests
        tplot_options("title", "")

    def test_markers_and_symbols_error_bars(self):
        # Regression test for lineplot crash when marker sizes are set
        # Taken from pytplot markers and symbols notebook in pyspedas_examples
        del_data("*")
        store_data(
            "data",
            data={"x": [1, 2, 3, 4, 5, 6], "y": [1, 1, 1, 1, 1, 1], "dy": [0.25] * 6},
        )
        timespan("1970-01-01", 10, "seconds")
        tplot_options("title", "Line plot with error bars")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "simple_lineplot_errbars.png"))
        options("data", "marker", "X")
        tplot_options("title", "Line plot with error bars and X markers")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "markers_lineplot_errbars.png"))
        options("data", "line_style", "None")
        tplot_options("title", "X markers and error bars with no lines")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "symbols_lineplot_errbars.png"))
        options("data", "marker_size", 30)
        tplot_options("title", "No lines, size 30 X markers and error bars")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "markersize_lineplot_errbars.png"))
        options("data", "marker", None)
        options("data", "line_style", "solid")
        tplot_options("title", "Line plot without markers with error bars")
        tplot(
            "data",
            display=global_display,
            save_png=os.path.join(save_dir, "markersize_nosymbols_lineplot_errbars.png"),
        )
        options("data", "marker_size", 20)
        options("data", "line_style", "None")
        options("data", "symbols", False)
        tplot_options("title", "Error bars, no lines, with size 20 markers but symbols=False")
        tplot(
            "data",
            display=global_display,
            save_png=os.path.join(save_dir, "markersize20_nosymbols_lineplot_errbars.png"),
        )
        options("data", "markevery", 2)
        options("data", "marker", "X")
        options("data", "line_style", "solid")
        tplot_options("title", "Line plot error bars and markers every 2 data points")
        options("data", "markevery", 2)
        options("data", "errorevery", 3)
        options("data", "marker", "X")
        options("data", "line_style", "solid")
        tplot_options("title", "Line plot error bars every 3 and markers every 2 data points")
        tplot(
            "data",
            display=global_display,
            save_png=os.path.join(save_dir, "markevery_errorevery_lineplot_errbars.png"),
        )
        options("data", "marker", "H")
        tplot_options("title", "Line plot with error bars and hexagon markers")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "hexagons_lineplot_errbars.png"))
        options("data", "capsize", 10)
        options("data", "ecolor", "r")
        options("data", "elinewidth", 5)
        tplot_options("title", "Line plot with red error bars and hexagon markers, capsize=10")
        tplot(
            "data",
            display=global_display,
            save_png=os.path.join(save_dir, "hexagons_lineplot_errbars_cap10redwidth5.png"),
        )
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests
        tplot_options("title", "")

    def test_plot_titles(self):
        # Regression test for lineplot crash when marker sizes are set
        # Taken from pytplot markers and symbols notebook in pyspedas_examples
        del_data("*")
        store_data("data", data={"x": [1, 2, 3, 4, 5, 6], "y": [1, 1, 1, 1, 1, 1]})
        timespan("1970-01-01", 10, "seconds")
        tplot_options("title", "This title should be visible")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "simple_title.png"))
        tplot_options("title", "This title should be in a larger font")
        tplot_options("title_size", 20)
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "title_titlesize.png"))
        tplot_options("title_size", 12)
        tplot_options("title", "")
        options("data", "xtitle", "There should be no main title displayed. char_size=20")
        options("data", "xsubtitle", "this is an x subtitle")
        options("data", "xtitle_color", "red")
        options("data", "ytitle", "This is a ytitle")
        options("data", "ysubtitle", "this is a y subtitle")
        options("data", "ytitle_color", "green")
        options("data", "char_size", 20)
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "title_emptystring.png"))
        tplot_options("title", None)
        options("data", "xtitle", "There should be no main title displayed")
        tplot("data", display=global_display, save_png=os.path.join(save_dir, "title_none.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests

    def test_time_data_bars(self):
        del_data("*")
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests
        themis.fgm(probe="c", trange=default_trange)
        # use timebar to place a databar
        timebar(t=10000.0, varname="thc_fge_dsl", databar=True, color="black")
        # use databar to place a databar
        databar("thc_fgs_dsl", -10000.0, color="red", dash=True)
        timebar(t=-10000.0, varname="thc_fge_dsl", databar=True, color="red", dash=True)
        timebar(
            t=time_double("2007-03-23/14:00"),
            varname="thc_fge_btotal",
            color="magenta",
        )
        timebar(t=time_double("2007-03-23/14:30"), color="blue")
        timebar(t=time_double("2007-03-23/15:30"), color="green", dash=True)
        tplot_options(
            "title",
            "Databars at +/- 10000 top panel, timebars at 14:30 and 15:30 all panels, timebar at 14:00 bottom panel, linestyles and colors as specified",
        )
        tplot(
            ["thc_fge_dsl", "thc_fge_btotal"],
            display=global_display,
            save_png=os.path.join(save_dir, "timebars.png"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests

    def test_line_pseudovariables(self):
        del_data("*")
        # Test that tplot variables with different number of traces can be combined into a pseudovariable and plotted correctly.
        # Both plots should have 4 properly labeled traces.
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests
        store_data("comb_3_1", data=["thc_fge_dsl", "thc_fge_btotal"])
        store_data("comb_1_3", data=["thc_fge_btotal", "thc_fge_dsl"])
        tplot_options("title", "Pseudovariable with one+three line traces")
        tplot("comb_1_3", display=global_display, save_png=os.path.join(save_dir, "pseudovars_comb_1_3.png"))
        tplot_options("title", "Pseudovariable with three+one line traces")
        tplot("comb_3_1", display=global_display, save_png=os.path.join(save_dir, "pseudovars_comb_3_1.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests

    def test_pseudovar_color_options(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        store_data("test_pseudo_colors", data=["thc_fge_dsl", "thc_fge_btotal"])
        # Set the color option on the pseudovariable (4 traces total, so 4 colors)
        options("test_pseudo_colors", "color", ["k", "r", "g", "b"])
        tplot_options("title", "Trace colors should be black, red, green, blue")
        tplot(
            "test_pseudo_colors",
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_colors_4color.png"),
        )  # should plot without "incorrect number of line colors" messages
        options("test_pseudo_colors", "color", ["k"])  # All black
        tplot_options("title", "Trace colors all black")
        tplot(
            "test_pseudo_colors",
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_colors_allsamecolor.png"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_var_line_options(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        options("thc_fgs_dsl", "line_style", ["solid", "dot", "dash"])
        tplot_options("title", "Line styles solid, dot, dash")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_linestyle_3styles.png"))
        options("thc_fgs_dsl", "line_style", "dot")  # gets used for all lines
        tplot_options("title", "Line styles all dot")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_linestyle_allsame.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_ytick_options(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        tplot_options(
            "title",
            "Default major and minor ticks, major ticks every 5000 nT from -5000 to 20000",
        )
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "ticks_default.png"))
        options(
            "thc_fgs_dsl",
            "y_major_ticks",
            [
                -7500,
                -5000,
                -2500,
                0,
                2500,
                5000,
                7500,
                10000,
                12500,
                15000,
                17500,
                20000,
                22500,
                25000,
            ],
        )
        tplot_options("title", "Manual major ticks, every 2500 nT from -7500 to 25000")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "ticks_major_every2500.png"))
        options("thc_fgs_dsl", "y_minor_tick_interval", 1250)
        tplot_options(
            "title",
            "Manual major and minor ticks, every 2500/1250 nT from -7500 to 25000",
        )
        tplot(
            "thc_fgs_dsl",
            display=global_display,
            save_png=os.path.join(save_dir, "ticks_major_every2500_minor1250.png"),
        )

        tplot_options(
            "title",
            "Y tick options: length 20, tick color green, width 5, labelcolor green, direction inout",
        )
        options("thc_fgs_gse", "ytick_length", 20)
        options("thc_fgs_gse", "ytick_color", "green")
        options("thc_fgs_gse", "ytick_width", 5)
        options("thc_fgs_gse", "ytick_direction", "inout")
        options("thc_fgs_gse", "ytick_labelcolor", "green")
        tplot("thc_fgs_gse", display=global_display, save_png=os.path.join(save_dir, "ytick_multi_options.png"))

        options("thc_fgs_gse", "ytick_length", None)
        options("thc_fgs_gse", "ytick_color", None)
        options("thc_fgs_gse", "ytick_width", None)
        options("thc_fgs_gse", "ytick_direction", None)
        options("thc_fgs_gse", "ytick_labelcolor", None)
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        tplot_options("title", "")

    def test_xtick_options(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")

        tplot_options(
            "title",
            "X tick options: length 20, tick color green, width 5, labelcolor green, direction inout",
        )
        options("thc_fgs_gse", "xtick_length", 20)
        options("thc_fgs_gse", "xtick_color", "green")
        options("thc_fgs_gse", "xtick_width", 5)
        options("thc_fgs_gse", "xtick_direction", "inout")
        options("thc_fgs_gse", "xtick_labelcolor", "green")
        tplot("thc_fgs_gse", display=global_display, save_png=os.path.join(save_dir, "xtick_multi_options.png"))
        options("thc_fgs_gse", "xtick_length", None)
        options("thc_fgs_gse", "xtick_color", None)
        options("thc_fgs_gse", "xtick_width", None)
        options("thc_fgs_gse", "xtick_direction", None)
        options("thc_fgs_gse", "xtick_labelcolor", None)
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        tplot_options("title", "")

    def test_ylim(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        ylim("thc_fgs_dsl", -100, 100)
        tplot_options("title", "Y limit [-100, 100]")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_ylim.png"))
        ylim("thc_fgs_dsl", 1, 100, 1)
        tplot_options("title", "Y limit [1, 100, 1] (should be log scaled)")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_ylim_1_100_log.png"))
        ylim("thc_fgs_dsl", 1, 200)
        tplot_options("title", "Y limit [1, 200] (should still be log scaled)")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_ylim_1_200_log.png"))
        ylim("thc_fgs_dsl", 1, 200, logflag=False)
        tplot_options("title", "Y limit [1, 200, False] (should now be linear scale)")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_ylim_1_200_linear.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_highlight(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        t1 = time_double("2007-03-23/03:00:00")
        t2 = time_double("2007-03-23/04:00:00")
        t3 = time_double("2007-03-23/05:00:00")
        highlight("thc_fgs_dsl", range=[t1, t2], color="red", alpha=0.2)
        highlight("thc_fgs_*", range=[t2, t3], color="green", alpha=0.2, hatch="o")
        tplot(
            "thc_fgs_dsl thc_fgs_gsm",
            display=global_display,
            save_png=os.path.join(save_dir, "test_highlight.png"),
        )

    def test_annotations(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        t2 = np.datetime64("2007-03-23T18:00")
        y2 = 10000
        # Position relative to panel size
        annotate(
            "thc_fgs_dsl",
            text="Annotation 1",
            position=[0.5, 0.5],
            color="red",
            fontsize=10,
        )
        # Position in data coordinates
        annotate(
            "thc_fgs_gsm",
            text="Annotation 2",
            position=(t2, y2),
            xycoords="data",
            color="blue",
            fontsize=20,
        )
        # Multi-variable annotations using wildcard
        annotate("thc_fgs_*", text="Multi Annotation", position=[0.1, 0.8], color="green")
        tplot(
            "thc_fgs_dsl thc_fgs_gsm",
            display=global_display,
            save_png=os.path.join(save_dir, "test_annotations.png"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_is_pseudovar(self):
        del_data("*")

        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        store_data("test_pseudo_colors", data=["thc_fge_dsl", "thc_fge_btotal"])
        self.assertTrue(is_pseudovariable("test_pseudo_colors"))
        self.assertFalse(is_pseudovariable("thc_fgs_dsl"))
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_xytitles(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        options("thc_fgs_dsl", "xtitle", "thc_fgs_dsl xtitle")
        options("thc_fgs_gsm", "xtitle", "thc_fgs_gsm xtitle")
        options("thc_fgs_dsl", "xsubtitle", "thc_fgs_dsl xsubtitle")
        options("thc_fgs_gsm", "xsubtitle", "thc_fgs_gsm xsubtitle")
        options("thc_fgs_dsl", "ytitle", "thc_fgs_dsl ytitle")
        options("thc_fgs_gsm", "ytitle", "thc_fgs_gsm ytitle")
        options("thc_fgs_dsl", "ysubtitle", "thc_fgs_dsl ysubtitle")
        options("thc_fgs_gsm", "ysubtitle", "thc_fgs_gsm ysubtitle")
        tplot_options("title", "xtitle and xsubtitle should show below each variable panel")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_titles_single.png"))
        tplot(
            "thc_fgs_dsl thc_fgs_gsm",
            display=global_display,
            save_png=os.path.join(save_dir, "test_titles_multi.png"),
        )
        tplot_options("title", "")

    def test_legend(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        options("thc_fgs_dsl", "marker", "x")
        options("thc_fgs_dsl", "legend_names", ["X-DSL", "Y-DSL", "Z-DSL"])
        options("thc_fgs_dsl", "legend_title", "legend title")
        options("thc_fgs_dsl", "legend_titlesize", 20)
        options("thc_fgs_dsl", "legend_color", "green")
        options("thc_fgs_dsl", "legend_edgecolor", "blue")
        options("thc_fgs_dsl", "legend_facecolor", "gray")
        options("thc_fgs_dsl", "legend_markerscale", 2)
        options("thc_fgs_dsl", "legend_markerfirst", True)
        options("thc_fgs_dsl", "legend_linewidth", 6)
        options("thc_fgs_dsl", "legend_ncols", 3)
        options("thc_fgs_dsl", "legend_shadow", True)
        options("thc_fgs_dsl", "legend_frameon", True)
        options("thc_fgs_dsl", "legend_location", "spedas")
        tplot("thc_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "test_legend.png"))

    def test_count_traces(self):
        del_data("*")

        themis.fgm(probe="c", trange=default_trange)
        store_data("test_pseudo_colors", data=["thc_fge_dsl", "thc_fge_btotal"])
        tr_fge = count_traces("thc_fge_dsl")
        tr_btotal = count_traces("thc_fge_btotal")
        tr_pseudo = count_traces("test_pseudo_colors")
        self.assertEqual(tr_fge, 3)
        self.assertEqual(tr_btotal, 1)
        self.assertEqual(tr_pseudo, 4)
        # We need to ensure thc_fge_dsl has a 'v' component, so we can pretend this variable is a spectrogram
        fgs_data = get_data("thc_fge_dsl")
        fgs_meta = get_data("thc_fge_dsl", metadata=True)
        self.assertTrue(fgs_meta is not None)
        store_data("thc_fge_dsl", data={"x": fgs_data.times, "y": fgs_data.y, "v": [0, 1, 2]})
        options("thc_fge_dsl", "spec", 1)
        tr_pseudo_spec = count_traces("test_pseudo_colors")
        self.assertEqual(tr_pseudo_spec, 1)
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_pseudovar_line_options(self):
        del_data("*")
        themis.fgm(probe="c", trange=default_trange)
        timespan("2007-03-23", 1, "days")
        store_data("test_pseudo_lineopts", data=["thc_fge_dsl", "thc_fge_btotal"])
        # Set the line_style on the pseudovariable (4 traces total, so 4 styles)
        options("test_pseudo_lineopts", "line_style", ["dot", "dash", "solid", "dash_dot"])
        tplot_options("title", "Pseudovar line styles dot, dash, solid, dash_dot")
        tplot(
            "test_pseudo_lineopts",
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_lineopts_4style.png"),
        )
        # Set all traces to the same style
        options("test_pseudo_lineopts", "line_style", "dot")
        tplot_options("title", "Pseudovar line styles all dot")
        tplot(
            "test_pseudo_lineopts",
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_lineopts_allsame"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_specplot_optimizations(self):
        del_data("*")
        ask_vars = themis.ask(trange=["2013-11-05", "2013-11-06"])
        self.assertTrue(len(ask_vars) > 0)
        timespan("2013-11-05", 1, "days")
        # Should plot without errors, show something other than all-blue or vertical lines
        tplot_options("title", "Should be mostly dark with a few lighter features")
        tplot(["thg_ask_atha"], display=global_display, save_png=os.path.join(save_dir, "thg_ask_atha.png"))
        options("thg_ask_atha", "y_no_resample", 1)
        tplot_options("title", "Should be mostly dark with a few lighter features")
        tplot("thg_ask_atha", display=global_display, save_png=os.path.join(save_dir, "thg_ask_atha_no_resample.png"))
        tplot_options("title", "Should be somewhat lighter now with logarithmic z scale")
        options("thg_ask_atha", "zlog", "log")
        tplot(
            "thg_ask_atha",
            display=global_display,
            save_png=os.path.join(save_dir, "thg_ask_atha_no_resample_zlog.png"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_themis_peef_bins(self):
        del_data("*")
        themis.esa(probe="a", trange=["2016-12-11", "2016-12-12"])
        timespan("2016-12-11", 1, "days")
        options("tha_peef_en_eflux", "yrange", [1000, 3000])
        tplot_options(
            "title",
            "Logarithmic bin boundaries: should be a boundary at y~=1947 eV, just below the tick mark at y=2000",
        )
        tplot("tha_peef_en_eflux", display=global_display, save_png=os.path.join(save_dir, "tha_peef_en_eflux.png"))
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_themis_peef_bins_interp_x(self):
        del_data("*")
        themis.esa(probe="a", trange=["2016-12-11", "2016-12-12"])
        timespan("2016-12-11", 1, "hours")
        options("tha_peef_en_eflux", "yrange", [1000, 3000])
        options("tha_peef_en_eflux", "x_interp", 1)
        options("tha_peef_en_eflux", "x_interp_points", 500)
        tplot_options("title", "Interpolated along time axis")
        tplot(
            "tha_peef_en_eflux",
            display=global_display,
            save_png=os.path.join(save_dir, "tha_peef_en_eflux_interp_x.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        options("tha_peef_en_eflux", "x_interp", 0)  # reset for other tests

    def test_themis_peef_bins_interp_y(self):
        del_data("*")
        themis.esa(probe="a", trange=["2016-12-11", "2016-12-12"])
        timespan("2016-12-11", 1, "hours")
        options("tha_peef_en_eflux", "yrange", [1000, 3000])
        options("tha_peef_en_eflux", "y_interp", 1)
        options("tha_peef_en_eflux", "y_interp_points", 200)
        tplot_options("title", "Interpolated along y axis")
        tplot(
            "tha_peef_en_eflux",
            display=global_display,
            save_png=os.path.join(save_dir, "tha_peef_en_eflux_interp_y.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        options("tha_peef_en_eflux", "y_interp", 0)  # reset for other tests

    def test_themis_peef_bins_interp_both(self):
        del_data("*")
        themis.esa(probe="a", trange=["2016-12-11", "2016-12-12"])
        timespan("2016-12-11", 1, "hours")
        options("tha_peef_en_eflux", "yrange", [1000, 3000])
        options("tha_peef_en_eflux", "x_interp", 1)
        options("tha_peef_en_eflux", "x_interp_points", 500)
        options("tha_peef_en_eflux", "y_interp", 1)
        options("tha_peef_en_eflux", "y_interp_points", 200)
        tplot_options("title", "Interpolated along both x and y axes")
        tplot(
            "tha_peef_en_eflux",
            display=global_display,
            save_png=os.path.join(save_dir, "tha_peef_en_eflux_interp_both.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        options("tha_peef_en_eflux", "x_interp", 0)  # reset for other tests
        options("tha_peef_en_eflux", "y_interp", 0)  # reset for other tests

    def test_mms_epsd_specplot(self):
        del_data("*")

        timespan("2015-08-01", 1, "days")
        # Logarithmic Y scale with lowest bin boundary = 0.0 by linear extrapolation from bin centers
        data = mms.mms_load_dsp(
            trange=["2015-08-01", "2015-08-02"],
            datatype=["epsd", "bpsd"],
            level="l2",
            data_rate="fast",
        )
        self.assertTrue("mms1_dsp_epsd_omni" in data)
        # options('mms1_dsp_epsd_omni','yrange',[8.0,130000.0])
        tplot(
            ["mms1_dsp_epsd_omni", "mms1_dsp_bpsd_omni"],
            display=global_display,
            save_png=os.path.join(save_dir, "mms1_epsd_omni.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_elfin_specplot(self):
        del_data("*")

        # ELFIN data with V values that oscillate, the original problem that resulted in the resample, this is an angular distrubtion
        timespan("2021-07-14/11:55", 10, "minutes")
        epd_var = elfin.epd(
            trange=["2021-07-14/11:55", "2021-07-14/12:05"],
            probe="a",
            level="l2",
            type_="nflux",
            fullspin=False,
        )
        self.assertTrue("ela_pef_hs_nflux_ch0" in epd_var)
        tplot_options("title", "ELFIN data with time-varying bins, should render accurately")
        tplot("ela_pef_hs_nflux_ch0", display=global_display, save_png=os.path.join(save_dir, "ELFIN_test.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests

    def test_fast_specplot(self):
        del_data("*")

        # FAST TEAMS has fill values -1e31 in V, top is an energy distribution, the bottom two are pitch angle distributions
        teams_vars = fast.teams(["1998-09-05", "1998-09-06"], level="k0")
        self.assertTrue(len(teams_vars) > 0)
        timespan("1998-09-05", 1, "days")
        tplot_options("title", "Fill should be removed, bottom two panels should go to Y=-90 deg")
        # Specify variables to plot as a space-delimited string with wildcards
        tplot("*H+ *H+_low *H+_high", display=global_display, save_png=os.path.join(save_dir, "TEAMS_test.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # reset to avoid interfering with other tests

    def test_themis_esa_specplot(self):
        del_data("*")

        # THEMIS ESA has monotonically decreasing energies, time varying energies, and also has fill
        esa_vars = themis.esa(trange=["2016-07-23", "2016-07-24"], probe="a")
        self.assertTrue(len(esa_vars) > 0)
        timespan("2016-07-23", 1, "days")
        tplot_options(
            "title",
            "Decreasing and time-varying energies, fillvals, should render correctly",
        )
        tplot("tha_peef_en_eflux", display=global_display, save_png=os.path.join(save_dir, "PEEF_test.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_themis_esa_copy_specplot(self):
        del_data("*")

        # THEMIS ESA has monotonically decreasing energies, time varying energies, and also has fill
        esa_vars = themis.esa(trange=["2016-07-23", "2016-07-24"], probe="a")
        self.assertTrue(len(esa_vars) > 0)
        timespan("2016-07-23", 1, "days")
        tplot_options(
            "title",
            "Decreasing and time-varying energies, fillvals, should render correctly",
        )
        tplot_copy("tha_peef_en_eflux", new_name="tha_peef_copy")
        tplot("tha_peef_copy", display=global_display, save_png=os.path.join(save_dir, "PEEF_copy_test.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_erg_specplot(self):
        del_data("*")

        # ERG specplots, only vertical lines on the bottom panel for original resample...
        erg.hep(trange=["2017-03-27", "2017-03-28"])
        timespan("2017-03-27", 1, "days")
        tplot_options("title", "Time varying spectral bins, should render correctly")
        tplot(
            ["erg_hep_l2_FEDO_L", "erg_hep_l2_FEDO_H"],
            display=global_display,
            save_png=os.path.join(save_dir, "ERG_test.png"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_maven_specplot(self):
        del_data("*")

        sta_vars = maven.spdf.load(
            trange=["2020-12-30", "2020-12-31"],
            instrument="static",
            datatype="c0-64e2m",
        )
        print(sta_vars)
        timespan("2020-12-30", 1, "days")
        # This variable contains all zeroes, and is set to plot with log scaling
        tplot_options("title", "Should be all the same color")
        tplot("bkg", display=global_display, save_png=os.path.join(save_dir, "MAVEN_test.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    # @unittest.skip(reason="Failing until we establish a default for spec_dim_to_plot")
    def test_maven_fluxes_specplot(self):
        del_data("*")

        swe_vars = maven.spdf.load(trange=["2014-10-18", "2014-10-19"], instrument="swea")
        print(swe_vars)
        timespan("2014-10-18", 1, "days")
        # This variable has 3 dimensions but is not marked in the CDF as being a specplot.
        # This used to crash in reduce_spec_dataset because the spec_dim_to_plot option was missing.
        tplot_options("title", "Spec data plotted as lines")
        tplot(
            "diff_en_fluxes",
            display=global_display,
            save_png=os.path.join(save_dir, "MAVEN_fluxes_test_nospec.png"),
        )
        options("diff_en_fluxes", "spec", 1)
        # Setting the "spec" option also sets the spec_dim_to_plot option to v2 in this case
        tplot_options("title", "Plotting as spectrum with default spec_dim_to_plot (v2)")
        tplot("diff_en_fluxes", display=global_display, save_png=os.path.join(save_dir, "MAVEN_fluxes_test_v2.png"))
        # Test that the "v1" option also works (it used to crash looking for "v" and not checking "v1")
        options("diff_en_fluxes", "spec_dim_to_plot", "v1")
        tplot_options("title", "Plotting as spectrum with spec_dim_to_plot=v1")
        tplot("diff_en_fluxes", display=global_display, save_png=os.path.join(save_dir, "MAVEN_fluxes_test_v1.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_pseudovars_title(self):
        del_data("*")

        themis.state(probe="c", trange=default_trange)
        store_data("ps1", ["thc_spin_initial_delta_phi", "thc_spin_idpu_spinper"])
        store_data("ps2", ["thc_spin_initial_delta_phi", "thc_spin_idpu_spinper"])
        store_data("ps3", ["thc_spin_initial_delta_phi", "thc_spin_idpu_spinper"])
        tplot_options("title", "Plot title should only appear once at top of plot")
        plotvars = ["thc_pos", "ps1", "thc_vel", "ps2", "thc_pos_gse", "ps3"]
        # Should have only one title at the top of the plot
        tplot(plotvars, save_png=os.path.join(save_dir, "test_pseudovars_title.png"), display=global_display)
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_pseudovars_spectra(self):
        del_data("*")

        # Load ESA and SST data
        themis.esa(probe="a", trange=default_trange)
        themis.sst(probe="a", trange=default_trange)

        # Make a combined variable with both ESA and SST spectral data (disjoint energy ranges)
        store_data("combined_spec", ["tha_peif_en_eflux", "tha_psif_en_eflux"])
        options("tha_peif_en_eflux", "y_no_resample", 1)
        # options('combined_spec','y_range',[5.0, 7e+06])
        vars = ["tha_psif_en_eflux", "combined_spec", "tha_peif_en_eflux"]
        tplot_options(
            "title",
            "Pseudovar with two spectra, disjoint energies: top=SST, middle=combined, bottom=ESA\nMiddle plot y range should go down to near zero",
        )
        tplot(
            vars,
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_spectra_disjoint_energies.png"),
        )

        tplot_options("title", "Original burst variable")
        tplot("tha_peib_en_eflux", display=global_display)
        degap("tha_peib_en_eflux", dt=4.0)
        tplot_options("title", "Degapped burst variable")
        tplot("tha_peib_en_eflux", display=global_display)
        # Make a combined variable with full & burst data (same energy ranges, intermittent burst data at higher cadence)
        store_data("esa_srvy_burst", ["tha_peif_en_eflux", "tha_peib_en_eflux"])
        # zlim('tha_peif_en_eflux', 1.0e3, 1.0e7)
        # options('tha_peif_en_eflux', 'y_range', [0.5, 1.0e6])

        options("tha_peib_en_eflux", "y_no_resample", 1)
        # zlim('tha_peib_en_eflux', 1.0e3, 1.0e7)
        # options('tha_peib_en_eflux', 'y_range', [0.5, 1.0e6])
        options("tha_peib_en_eflux", "data_gap", 4.0)
        vars = ["tha_peif_en_eflux", "esa_srvy_burst", "tha_peib_en_eflux"]
        tplot_options(
            "title",
            "Combining full & burst cadence with same energies: top=fast, middle=combined, bot=burst",
        )
        tplot(vars, display=global_display, save_png=os.path.join(save_dir, "test_pseudo_spectra_full_burst.png"))
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

        # Zoom in o a burst interval
        # timespan('2007-03-23/12:20', 10, 'minutes')
        tplot_options(
            "title",
            "Combined full and burst cadence with same energies (zoomed in) top=fast, mid=combined, bot=burst",
        )
        tplot(
            vars,
            trange=["2007-03-23/12:20", "2007-03-23/12:30"],
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_spectra_full_burst_zoomed.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        tplot_options("title", "")

    def test_pseudovars_spectra_explicit_yzranges(self):
        del_data("*")

        # Load ESA and SST data
        themis.esa(probe="a", trange=default_trange)
        themis.sst(probe="a", trange=default_trange)

        # Make a combined variable with both ESA and SST spectral data (disjoint energy ranges)
        store_data("combined_spec", ["tha_peif_en_eflux", "tha_psif_en_eflux"])
        options("tha_peif_en_eflux", "y_no_resample", 1)
        options("combined_spec", "y_range", [0.1, 7e06])
        vars = ["tha_psif_en_eflux", "combined_spec", "tha_peif_en_eflux"]
        tplot_options(
            "title",
            "Pseudovar with two spectra, disjoint energies: top=SST, middle=combined, bottom=ESA\nMiddle plot y range should go down to 0.1",
        )
        tplot(
            vars,
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_spectra_disjoint_energies_explicityrange.png"),
        )

        tplot_options(
            "title",
            "Pseudovar with two spectra, disjoint energies: top=SST, middle=combined, bottom=ESA\nMiddle plot should have linear Y axis",
        )
        options("combined_spec", "ylog", False)
        tplot(
            vars,
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_spectra_disjoint_energies_explicityrange_linear.png"),
        )

        store_data("esa_srvy_burst", ["tha_peif_en_eflux", "tha_peib_en_eflux"])
        options("tha_peib_en_eflux", "y_no_resample", 1)
        zlim("tha_peib_en_eflux", 1.0e3, 1.0e7)
        options("tha_peib_en_eflux", "y_range", [0.5, 1.0e6])
        options("tha_peib_en_eflux", "data_gap", 4.0)
        vars = ["tha_peif_en_eflux", "esa_srvy_burst", "tha_peib_en_eflux"]
        zlim("esa_srvy_burst", 1.0e3, 1.0e20)
        tplot_options(
            "title",
            "Combining full & burst cadence with same energies: top=fast, middle=combined, bot=burst\nMiddle panel should have an expanded zrange (high max)",
        )
        tplot(
            vars,
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_spectra_full_burst_explicitzrange.png"),
        )
        zlim("esa_srvy_burst", 1.0e5, 1.0e6, False)
        tplot_options(
            "title",
            "Combining full & burst cadence with same energies: top=fast, middle=combined, bot=burst\nMiddle panel should have a reduced zrange with linear scaling",
        )
        tplot(
            vars,
            display=global_display,
            save_png=os.path.join(save_dir, "test_pseudo_spectra_full_burst_explicitzrange_zlin.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        tplot_options("title", "")

    def test_pseudo_spectra_plus_line(self):
        del_data("*")

        mms.fpi(datatype="des-moms", trange=["2015-10-16", "2015-10-17"])
        mms.edp(trange=["2015-10-16", "2015-10-17"], datatype="scpot")
        # Create a pseudovariable with an energy spectrum plus a line plot of spacecraft potential
        store_data("spec", data=["mms1_des_energyspectr_omni_fast", "mms1_edp_scpot_fast_l2"])
        # Set some options so that the spectrum, trace, and y axes are legible
        options("mms1_edp_scpot_fast_l2", "yrange", [10, 100])
        options("mms1_edp_scpot_fast_l2", "alpha", 0.5)
        # options('mms2_edp_scpot_fast_l2', 'right_axis', True)
        options("spec", "right_axis", "True")
        options("mms1_des_energyspectr_omni_fast", "ztitle", "This is a ztitle")
        options("mms1_des_energyspectr_omni_fast", "zsubtitle", "This is a z subtitle")
        options("mms1_des_energyspectr_omni_fast", "ztitle_color", "green")
        options("mms1_des_energyspectr_omni_fast", "second_axis_size", 0.14)
        tplot_options("xmargin", [0.1, 0.2])
        timespan("2015-10-16", 1, "days")
        tplot_options(
            "title",
            "Pseudovar with energy spectrum plus line plot of s/c potential, combined var has right_axis set\nTop: spec Middle: combined Bottom: line",
        )
        tplot(
            "mms1_des_energyspectr_omni_fast spec mms1_edp_scpot_fast_l2",
            xsize=12,
            display=global_display,
            save_png=os.path.join(save_dir, "MMS_pseudo_spec_plus_line.png"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_pseudo_spectra_plus_line_copy(self):
        del_data("*")

        mms.fpi(datatype="des-moms", trange=["2015-10-16", "2015-10-17"])
        mms.edp(trange=["2015-10-16", "2015-10-17"], datatype="scpot")
        # Create a pseudovariable with an energy spectrum plus a line plot of spacecraft potential
        store_data("spec", data=["mms1_des_energyspectr_omni_fast", "mms1_edp_scpot_fast_l2"])
        # Set some options so that the spectrum, trace, and y axes are legible
        options("mms1_edp_scpot_fast_l2", "yrange", [10, 100])
        # options('mms2_edp_scpot_fast_l2', 'right_axis', True)
        options("spec", "right_axis", "True")
        tplot_options("xmargin", [0.1, 0.2])
        timespan("2015-10-16", 1, "days")
        tplot_options("title", "Pseudovar with energy spectrum plus line plot of s/c potential")
        tplot_copy("spec", "spec_copy")
        tplot(
            "spec_copy",
            xsize=12,
            display=global_display,
            save_png=os.path.join(save_dir, "MMS_pseudo_spec_plus_line.png"),
        )
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_psp_flux_plot(self):
        del_data("*")

        spi_vars = psp.spi(
            trange=["2022-12-12/00:00", "2022-12-12/23:59"],
            datatype="sf00_l3_mom",
            level="l3",
            time_clip=True,
        )
        self.assertTrue(len(spi_vars) > 0)
        d0 = get_data("psp_spi_EFLUX_VS_ENERGY")
        time = d0.times
        # print(time)

        d1 = get_data("psp_spi_EFLUX_VS_ENERGY")
        energy_channel = d1.v
        # print(energy_channel)

        d2 = get_data("psp_spi_EFLUX_VS_ENERGY")
        energy_flux = d2.y
        # print(energy_flux)
        # e_flux = data_quants["psp_spi_EFLUX_VS_ENERGY"].coords["v"].values
        energy_flux[energy_flux == 0] = np.nan
        store_data("E_Flux", data={"x": time.T, "y": energy_flux, "v": energy_channel})
        options("E_Flux", opt_dict={"Spec": 1, "zlog": 1, "Colormap": "jet", "ylog": 1})
        timespan("2022-12-12", 1, "days")
        tplot_options("title", "Parker Solar Probe E_flux")
        tplot("E_Flux", display=global_display, save_png=os.path.join(save_dir, "psp_E_Flux.png"))
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_save_with_plot_objects(self):
        # Test that output files are saved when return_plot_objects is True

        themis.state(probe="a")
        local_png = os.path.join(save_dir, "ret_plot_objs.png")
        if os.path.exists(local_png):
            os.remove(local_png)
        tplot_options("title", "Should create ret_plot_objs.png ")
        tplot(
            "tha_pos",
            display=global_display,
            return_plot_objects=True,
            save_png=local_png,
        )
        self.assertTrue(os.path.exists(local_png))
        tplot_options("title", "")

    def test_original_tplot_vlabels(self):
        # Test alternate varlabel implementation from Tomo Hori
        del_data("*")

        timespan("2017-03-27", 1, "days")
        erg.mgf()
        erg.orb()

        split_vec("erg_orb_l2_pos_rmlatmlt")
        split_vec("erg_orb_l2_pos_Lm")
        options("erg_orb_l2_pos_rmlatmlt_x", "ytitle", "R")
        options("erg_orb_l2_pos_rmlatmlt_y", "ytitle", "Mlat")
        options("erg_orb_l2_pos_rmlatmlt_z", "ytitle", "MLT")

        var_label = [
            "erg_orb_l2_pos_Lm_x",
            "erg_orb_l2_pos_rmlatmlt_x",
            "erg_orb_l2_pos_rmlatmlt_y",
            "erg_orb_l2_pos_rmlatmlt_z",
        ]
        # tplot_options('var_label', var_label)

        plot_vars = [
            "erg_mgf_l2_mag_8sec_sm",
            "erg_mgf_l2_igrf_8sec_sm",
            "erg_orb_l2_pos_Lm_x",
        ]

        tplot(
            plot_vars,
            var_label=var_label,
            display=global_display,
            save_png=os.path.join(save_dir, "original_varlabel.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        tplot_options("title", "")

    def test_tplot_vlabels_extra_panel(self):
        # Test alternate varlabel implementation from Tomo Hori
        del_data("*")
        timespan("2017-03-27", 1, "days")
        erg.mgf()
        erg.orb()

        split_vec("erg_orb_l2_pos_rmlatmlt")
        split_vec("erg_orb_l2_pos_Lm")
        options("erg_orb_l2_pos_rmlatmlt_x", "ytitle", "R")
        options("erg_orb_l2_pos_rmlatmlt_y", "ytitle", "Mlat")
        options("erg_orb_l2_pos_rmlatmlt_z", "ytitle", "MLT")

        var_label = [
            "erg_orb_l2_pos_Lm_x",
            "erg_orb_l2_pos_rmlatmlt_x",
            "erg_orb_l2_pos_rmlatmlt_y",
            "erg_orb_l2_pos_rmlatmlt_z",
        ]
        # tplot_options('var_label', var_label)
        tplot_options("varlabel_style", "extra_panel")
        plot_vars = [
            "erg_mgf_l2_mag_8sec_sm",
            "erg_mgf_l2_igrf_8sec_sm",
            "erg_orb_l2_pos_Lm_x",
        ]

        tplot(
            plot_vars,
            var_label=var_label,
            display=global_display,
            save_png=os.path.join(save_dir, "varlabel_extra_panel.png"),
        )

        options(var_label, "var_label_format", "{:.1f}")
        tplot_options("title", "Var labels should all have 1 digit after the decimal point")
        tplot(
            plot_vars,
            var_label=var_label,
            display=global_display,
            save_png=os.path.join(save_dir, "varlabel_extra_panel_single_decimal.png"),
        )
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests
        tplot_options("varlabel_style", None)
        tplot_options("title", "")
        timespan("2007-03-23", 1, "days")  # Reset to avoid interfering with other tests

    def test_tplot_trange(self):
        del_data("*")
        themis.fit(probe="e", trange=["2007-03-23", "2007-03-24"])
        tplot_options("title", "No timespan or xlim set, full time range of loaded data")
        tplot("the_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "notrange_default.png"))
        tplot_options("title", "No timespan or xlim set, using trange to plot 12:00 to 23:59")
        tplot(
            "the_fgs_dsl",
            trange=["2007-03-23/12:00", "2007-03-24/00:00"],
            display=global_display,
            save_png=os.path.join(save_dir, "trange_pm.png"),
        )
        tplot_options("title", "Using xlim to set time range from 00:00 to 12:00")
        xlim("2007-03-23/00:00:00", "2007-03-23/12:00:00")
        tplot("the_fgs_dsl", display=global_display, save_png=os.path.join(save_dir, "xlim_am.png"))
        tplot_options("title", "Xlim set to 00:00 to 12:00, overriding with trange from 10:00 to 12:00")
        tplot(
            "the_fgs_dsl",
            trange=["2007-03-23/10:00", "2007-03-23/12:00:00"],
            display=global_display,
            save_png=os.path.join(save_dir, "xlim_am_trange_override.png"),
        )
        tplot_options("title", "")


if __name__ == "__main__":
    unittest.main()



"""
Compute power spectra for a tplot variable.

Calls dpwrspc for the actual computation.

Notes
-----
Similar to tdpwrspc.pro in IDL SPEDAS.

"""
import logging
import numpy as np
from pyspedas.tplot_tools import dpwrspc
from pyspedas.tplot_tools import split_vec
from pyspedas.tplot_tools import get_data, store_data, options, time_double, set_units


def tdpwrspc(
    varname,
    newname=None,
    trange=["0.0", "0.0"],
    nboxpoints=None,
    nshiftpoints=None,
    polar=False,
    bin=3,
    nohanning=False,
    noline=False,
    notperhz=False,
    notmvariance=False,
):
    """
    Compute power spectra for a tplot variable.

    Parameters
    ----------
    varname: str
        Name of tplot variable.
    newname: str, optional
        Name of new tplot variable to save data to.
        Default: None. If newname is not set '_dpwrspc' will be appended to the varname
    trange: list of float, optional
        Start and end times for the calculation.
    nboxpoints: int, optional
        The number of points to use for the hanning window.
        Default: 256
    nshiftpoints: int, optional
        The number of points to shift for each spectrum.
        Default: 128
    polar: bool, optional
        If True, the input data is in polar coordinates.
        Default: False
    bin: int, optional
        Size for binning of the data along the frequency domain.
        Default: 3
    nohanning: bool, optional
        If True, no hanning window is applied to the input.
        Default: False
    noline: bool, optional
        If True, no straight line is subtracted.
        Default: False
    notperhz: bool, optional
        If True, the output units are the square of the input units.
        Default: False
    notmvariance: bool, optional
        If True, replace output spectrum for any windows that have variable.
        cadence with NaNs.
        Default: False

    Returns
    -------
    str
        Name of new tplot variable.

    Example
    -------
        >>> # Compute the power spectrum of a given time series
        >>> import pyspedas
        >>> import numpy as np
        >>> pyspedas.store_data('a', data={'x': range(3000), 'y': np.random.random(3000)})
        >>> pyspedas.tdpwrspc('a')
    """
    if newname is None:
        newname = varname + "_dpwrspc"

    data_tuple = get_data(varname)
    metadata = get_data(varname, metadata=True)

    # Find units
    inputunits = "#"
    has_data_att = False
    has_units = False
    newunits = ""
    if "data_att" in metadata:
        mm = metadata["data_att"]
        has_data_att = True
        if "units" in mm:
            inputunits = mm["units"]
            has_units = True

    if inputunits != "#":
        if notperhz:
            newunits = "(" + inputunits + ")^2"
        else:
            newunits = "(" + inputunits + ")^2/Hz"

    if data_tuple is not None:
        if data_tuple[1][0].shape != ():
            split_vars = split_vec(varname, polar=polar)
            out_vars = []
            for var in split_vars:
                out_vars.append(
                    tdpwrspc(
                        var,
                        newname=var + "_dpwrspc",
                        trange=trange,
                        nboxpoints=nboxpoints,
                        nshiftpoints=nshiftpoints,
                        polar=polar,
                        bin=bin,
                        nohanning=nohanning,
                        noline=noline,
                        notperhz=notperhz,
                        notmvariance=notmvariance,
                    )
                )
            return out_vars
        else:
            t = data_tuple[0]
            y = data_tuple[1]
            if trange is not None and len(trange) == 2 and trange[1] > trange[0]:
                # Only use points inside the time range
                tr = time_double(trange)
                ok = np.argwhere((t >= tr[0]) & (t < tr[1]))
                if len(ok) == 0:
                    logging.error("No data in time range")
                    logging.error(f"{tr}")
                    return None
                t = t[ok]
                y = y[ok]

            # Filter out NaNs
            ok = np.isfinite(y)
            if len(ok) == 0:
                logging.error("No finite data in time range")
                return None
            t = t[ok]
            y = y[ok]

            t00 = data_tuple[0][0]
            t = t - t00

            # Only do this if there are enough data points, default nboxpoints to
            # 64 and nshiftpoints to 32, and use larger values when there are more
            # points
            if nboxpoints is None:
                nbp = np.max([2 ** (np.floor(np.log(len(ok)) / np.log(2)) - 5), 8])
            else:
                nbp = nboxpoints

            if nshiftpoints is None:
                nsp = nbp / 2.0
            else:
                nsp = nshiftpoints

            if len(ok) <= nbp:
                logging.error("Not enough data in time range")
                return None

            pwrspc = dpwrspc(
                t,
                y,
                nboxpoints=nbp,
                nshiftpoints=nsp,
                bin=bin,
                nohanning=nohanning,
                noline=noline,
                notperhz=notperhz,
                notmvariance=notmvariance,
            )

            if pwrspc is not None:
                store_data(
                    newname, data={"x": pwrspc[0] + t00, "y": pwrspc[2], "v": pwrspc[1]}
                )
                options(newname, "ysubtitle", "[Hz]")
                options(newname, "spec", True)
                options(newname, "ylog", True)
                options(newname, "zlog", True)
                options(newname, "Colormap", "spedas")
                if has_units:
                    set_units(newname, newunits)
                if notperhz:
                    options(newname, "ztitle", "(" + inputunits + ")^2")
                else:
                    options(newname, "ztitle", "(" + inputunits + ")^2/Hz")

                return newname
    return None



import numpy as np
import pyspedas
from pyspedas.tplot_tools import get_data
import logging

def is_pseudovariable(tvar):
    """
    Checks if a tplot variable is a pseudovariable

    Parameter
    ----------
    tvar: str
        Name of tplot variable to check

    Return
    ----------
    bool:
        Return True if tplot variable is a pseudovariable.

    Example
    ----------
        >>> import pyspedas
        >>> pyspedas.store_data('a', data={'x': range(10), 'y': range(10)})
        >>> pyspedas.store_data('b', data={'x': range(10), 'y': range(10)})
        >>> pyspedas.store_data('pseudovar', data=['a','b'])
        >>> pyspedas.is_pseudovariable('a')  # False
        >>> pyspedas.is_pseudovariable('pseudo') # True

    """
    pseudo_var = False
    if tvar in pyspedas.tplot_tools.data_quants.keys():
        var_quants = pyspedas.tplot_tools.data_quants[tvar]

        if not isinstance(var_quants, dict):
            overplot_list = var_quants.attrs['plot_options'].get('overplots_mpl')
            if overplot_list is not None and len(overplot_list) > 0:
                pseudo_var = True

        var_data = get_data(tvar, dt=True)

        if isinstance(var_data, list) or isinstance(var_data, str) or pseudo_var:
            return True
    else:
        logging.warning("The name %s is not in pyspedas.",tvar)

    return False



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import numpy as np

def get_y_range(dataset):
    # This takes the data and sets the minimum and maximum range of the data values.
    # If the data type later gets set to 'spec', then we'll change the ymin and ymax
    import warnings
    warnings.filterwarnings("ignore")

    # Special rule if 'spec' is True
    if 'spec' in dataset.attrs['plot_options']['extras']:
        if dataset.attrs['plot_options']['extras']['spec']:
            try:
                ymin = np.nanmin(dataset.coords['spec_bins'].values)
                ymax = np.nanmax(dataset.coords['spec_bins'].values)
                return [ymin, ymax]
            except Exception as e:
                #continue on to the code below
                pass

    dataset_temp = dataset.where(dataset != np.inf)
    dataset_temp = dataset_temp.where(dataset != -np.inf)
    try:
        y_min = np.nanmin(dataset_temp.values)
        y_max = np.nanmax(dataset_temp.values)
    except RuntimeWarning:
        y_min = np.nan
        y_max = np.nan

    # CDF files may have array of strings (e.g., RBSP EMFISIS)
    if isinstance(y_min, str):
        y_min = np.nan
        y_max = np.nan
        
    if y_min == y_max:
        # Show 10% and 10% below the straight line
        y_min = y_min - (.1 * np.abs(y_min))
        y_max = y_max + (.1 * np.abs(y_max))
    warnings.resetwarnings()
    return [y_min, y_max]

